from .head_classifier import *

from .trainer_util import (
    export_onnx,
    model_save,
    export_onnx_optimization,
    export_onnx_quantization,
)

from .custom_ops import custom_scaled_dot_product_attention

from .image_classifier import ImageClassifierModule
from .text_classifier import TextClassifierModule
from .image_text_classifier import ImageTextClassifierModule



import pandas as pd
from torch.utils.data import Dataset
from typing import Optional
import torch
from pytorch_lightning import LightningDataModule
from typing import Union


class BaseDataModule(LightningDataModule):
    """
    Base DataModule for all classification tasks
    Args:
        train_dataset (pd.DataFrame): training dataset
        val_dataset (pd.DataFrame): validation dataset
        test_dataset (pd.DataFrame): test dataset
        train_config: train_config
    """

    def __init__(
        self,
        train_dataset: pd.DataFrame,
        val_dataset: pd.DataFrame,
        test_dataset: pd.DataFrame,
        train_config,
        feedback_dataset: Union[pd.DataFrame, None] = None,
    ):
        super().__init__()
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.feedback_dataset = feedback_dataset
        self.test_dataset = test_dataset
        self.batch_size = train_config.batch_size
        self.num_workers = train_config.num_workers
        self.problem_type = train_config.problem_type
        self.num_class = train_config.num_class

    def train_dataloader(self):
        dataloader_kwargs = {
            "num_workers": self.num_workers,
            "pin_memory": True,
            "drop_last": True,
            "shuffle": True,
        }
        dataloader_kwargs["batch_size"] = self.batch_size

        return torch.utils.data.DataLoader(self.trainset, **dataloader_kwargs)

    def val_dataloader(self):
        dataloader_kwargs = {"num_workers": self.num_workers, "pin_memory": True}
        dataloader_kwargs["batch_size"] = self.batch_size

        return torch.utils.data.DataLoader(self.valset, **dataloader_kwargs)

    def test_dataloader(self):
        dataloader_kwargs = {
            "num_workers": self.num_workers,
            "pin_memory": True,
        }
        dataloader_kwargs["batch_size"] = self.batch_size

        return torch.utils.data.DataLoader(self.testset, **dataloader_kwargs)



# from .pipeline import pipeline
import torch
import os
from .onnx_pipeline import (
    OnnxTextClassificationPipeline,
    OnnxImageClassificationPipeline,
    OnnxImageTextClassificationPipeline,
)
from .pytorch_pipeline import PytorchClassificationPipeline

# Register all supported pipeline for each model type.
MODEL_TYPE_PIPELINES = {
    "TextClassifier": {
        "pytorch": PytorchClassificationPipeline,
        "onnx": OnnxTextClassificationPipeline,
    },
    "ImageClassifier": {
        "pytorch": PytorchClassificationPipeline,
        "onnx": OnnxImageClassificationPipeline,
    },
    "ImageTextClassifier": {
        "pytorch": PytorchClassificationPipeline,
        "onnx": OnnxImageTextClassificationPipeline,
    },
}


def pipeline(
    model_type: str,
    save_dir: str,
    inference_engine: str = "onnx",
    device: str = "cpu",
    quantization: bool = False,
    optimization: bool = False,
):
    """
    Initialize pipeline for inference, support both onnx and pytorch inference engine.
    Each model type will have different pipeline, for example:
        - TextClassifier: OnnxTextClassificationPipeline, PytorchClassificationPipeline

    Args:
        model_type (str): Model type, support TextClassifier, ImageClassifier, ImageTextClassifier.
        save_dir (str): Model directory.
        inference_engine (str, optional): Inference engine, support onnx and pytorch. Defaults to "onnx".
        device (str, optional): Device for inference, support cpu and gpu. Defaults to "cpu".
    """
    assert os.path.isdir(save_dir), f"Not existing model directory {save_dir}"

    # Only support onnx and pytorch inference engine.
    if inference_engine not in ["onnx", "pytorch"]:
        raise ValueError(f"Not supported inference engine: {inference_engine}")

    if model_type not in MODEL_TYPE_PIPELINES.keys():
        raise ValueError(f"Not supported model type: {model_type}")

    pipeline_class = MODEL_TYPE_PIPELINES[model_type][inference_engine]

    return pipeline_class(model_type, save_dir, device, quantization, optimization)



import logging
import torch

from ..head_classifier import *
from .image_config import ImageClassifierConfig
from ..base_trainer import BaseClassifierModule
from ..base_feedback_trainer import FeedbackBaseClassifierModule

logger = logging.getLogger(__name__)


class ImageClassifierModule(BaseClassifierModule):
    """
    Image Classifier for training and inference.

    Args:
        train_config (ImageClassifierConfig): ImageClassifierConfig.
    """

    def __init__(self, train_config: ImageClassifierConfig):
        super().__init__(train_config)

        (
            self.backbone,
            self.train_transforms,
            self.val_transforms,
            self.data_config,
            self.train_config.timm,
            self.initial_config,
        ) = self.get_model()
        self.embedding_size = self.backbone.forward(
            torch.randn(self.data_config["input_size"]).unsqueeze(0)
        ).shape[-1]

        # Define classifier
        if type(self.train_config.fc_classifier) == str:
            classifier_class = globals()[self.train_config.fc_classifier]
            self.classifier_head = classifier_class(
                self.embedding_size, self.train_config.num_class
            )
        else:
            self.classifier_head = self.train_config.fc_classifier(
                self.embedding_size, self.train_config.num_class
            )

        self.save_hyperparameters()

    def get_model(self):
        """
        Get model, train transforms, val transforms, initial config, and timm flag.
        """
        return self.get_image_model(
            self.train_config.backbone_name,
            self.train_config.freeze,
            self.train_config.lora,
            self.train_config.lora_r,
            self.train_config.lora_alpha,
            self.train_config.lora_target_modules,
            self.train_config.lora_dropout,
        )

    # model forward
    def forward(
        self,
        input_img_paths,
        backbone=None,
        classifier_head=None,
        eval=False,
        classifier=True,
    ):
        if backbone is None:
            backbone = self.backbone
        if classifier_head is None:
            classifier_head = self.classifier_head
        return self.image_forward(
            input_img_paths,
            backbone,
            classifier_head,
            self.train_config.freeze,
            eval=eval,
            classifier=classifier,
        )

    def configure_optimizers(self):
        optimizer = self.classifier_head_optimizers()
        optimizer = self.add_optimizers(
            optimizer,
            self.backbone,
            self.train_config.freeze,
            self.train_config.lora,
            self.train_config.backbone_learning_rate,
        )

        lr_scheduler = self.get_lr_scheduler(optimizer, self.train_config.lr_scheduler)

        return [optimizer], [{"scheduler": lr_scheduler}]


class FeedbackImageClassifierModule(FeedbackBaseClassifierModule):
    """
    FeedbackImage Classifier for training and inference.

    Args:
        train_config (ImageClassifierConfig): ImageClassifierConfig.
    """

    def __init__(self, train_config: ImageClassifierConfig):
        super().__init__(train_config)

        (
            self.backbone,
            self.train_transforms,
            self.val_transforms,
            self.data_config,
            self.train_config.timm,
            self.initial_config,
        ) = self.get_model()
        self.embedding_size = self.backbone.forward(
            torch.randn(self.data_config["input_size"]).unsqueeze(0)
        ).shape[-1]

        # Define classifier
        if type(self.train_config.fc_classifier) == str:
            classifier_class = globals()[self.train_config.fc_classifier]
            self.classifier_head = classifier_class(
                self.embedding_size, self.train_config.num_class
            )
        else:
            self.classifier_head = self.train_config.fc_classifier(
                self.embedding_size, self.train_config.num_class
            )

        self.save_hyperparameters()

    def get_model(self):
        """
        Get model, train transforms, val transforms, initial config, and timm flag.
        """
        return self.get_image_model(
            self.train_config.backbone_name,
            self.train_config.freeze,
            self.train_config.lora,
            self.train_config.lora_r,
            self.train_config.lora_alpha,
            self.train_config.lora_target_modules,
            self.train_config.lora_dropout,
        )

    # model forward
    def forward(
        self,
        input_img_paths,
        backbone=None,
        classifier_head=None,
        eval=False,
        classifier=True,
    ):
        if backbone is None:
            if type(self.backbone) == nn.ModuleList:
                backbone = self.backbone[0]
            else:
                backbone = self.backbone
        if classifier_head is None:
            classifier_head = self.classifier_head
        return self.image_forward(
            input_img_paths,
            backbone,
            classifier_head,
            self.train_config.freeze,
            eval=eval,
            classifier=classifier,
        )

    def configure_optimizers(self):
        optimizer = self.classifier_head_optimizers()
        optimizer = self.add_optimizers(
            optimizer,
            self.backbone,
            self.train_config.freeze,
            self.train_config.lora,
            self.train_config.backbone_learning_rate,
        )

        lr_scheduler = self.get_lr_scheduler(optimizer, self.train_config.lr_scheduler)

        return [optimizer], [{"scheduler": lr_scheduler}]



import torch
import logging
from ..base_config import BaseClassifierConfig

logger = logging.getLogger(__name__)


class TextClassifierConfig(BaseClassifierConfig):
    """
    Text Classifier Config.

    Args:
        text_prompt (str): Text prompt for the task, default: "{}의 상품 종류는?".
        backbone_name (str): Name of the backbone model, default: "EleutherAI/polyglot-ko-1.3b".
        classifier_learning_rate (float): Learning rate for the classifier, default: 1e-3.
        backbone_learning_rate (float): Learning rate for the backbone, default: 3e-4.
        freeze (bool): Whether to freeze the backbone, default: False.
        lora (bool): Whether to use LoRA for the backbonㄷ, default: True.
        lora_r (int): LoRA r parameter, default: 16.
        lora_alpha (int): LoRA alpha parameter, default: 16.
        lora_target_modules (list): LoRA target modules, default: ["query_key_value"].
        lora_dropout (float): LoRA dropout, default: 0.1.
    """

    def __init__(
        self,
        text_prompt: str = "{}의 상품 종류는?",
        backbone_name: str = "EleutherAI/polyglot-ko-1.3b",
        classifier_learning_rate: float = 1e-3,
        backbone_learning_rate: float = 3e-4,
        freeze: bool = False,
        lora: bool = True,
        lora_r: int = 16,
        lora_alpha: int = 16,
        lora_target_modules: list = None,
        lora_dropout: float = 0.1,
        **kwargs
    ):
        super().__init__(**kwargs)
        # 태스크에 맞게 text prompt 수정 필수!
        self.text_prompt = text_prompt
        self.backbone_name = backbone_name
        self.classifier_learning_rate = classifier_learning_rate
        self.backbone_learning_rate = backbone_learning_rate

        self.freeze = freeze
        self.lora = lora
        if self.freeze and self.lora:
            raise ValueError(
                "freezing backbone and using lora cannot be True at the same time"
            )

        self.lora_r = lora_r
        self.lora_alpha = lora_alpha
        if lora_target_modules is None:
            self.lora_target_modules = ["query_key_value"]
        else:
            self.lora_target_modules = lora_target_modules
        self.lora_dropout = lora_dropout
        # Boolean to check if the model is loaded from transformers or timm, default value is False for TextClassifier
        self.timm = False



import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import torch

torch.backends.cudnn.benchmark = False
import logging
from typing import List, Optional, Union

import pandas as pd

from ..trainer_util import set_seed
from .text_config import TextClassifierConfig
from .text_datamodule import TextClassifierDataModule, FeedbackTextClassifierDataModule
from .text_trainer import TextClassifierModule, FeedbackTextClassifierModule
from ..base_classifier import BaseClassifier

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class TextClassifier(BaseClassifier):
    """
    Text Classifier for training and inference.

    Args:
        train_dataset (Union[pd.DataFrame, List[List]]): Training dataset.
        val_dataset (Union[pd.DataFrame, List[List]]): Validation dataset.
        test_dataset (Union[pd.DataFrame, List[List]]): Test dataset.
        label_name_list (List): List of label names.
        train_config (TextClassifierConfig): TextClassifierConfig.
    """

    def __init__(
        self,
        train_dataset: Union[pd.DataFrame, List[List]],
        val_dataset: Union[pd.DataFrame, List[List]],
        test_dataset: Union[pd.DataFrame, List[List]] = None,
        label_name_list: List = None,
        train_config: TextClassifierConfig = None,
    ):
        # Initial starting without loading from checkpoint
        if train_config is None:
            output_dir = "tmp_classifier"
            logger.info(
                f"No `TextClassifierConfig` passed, using `output_dir={output_dir}`."
            )
            train_config = TextClassifierConfig(output_dir=output_dir)
        else:
            if type(train_config) != TextClassifierConfig:
                raise ValueError(
                    f"train_config type has to be TextClassifierConfig, but got {type(train_config)}"
                )

        feedback_dataset = None
        super().__init__(
            train_dataset,
            val_dataset,
            feedback_dataset,
            test_dataset,
            label_name_list,
            train_config,
        )
        # Define Data module
        self.data_module = TextClassifierDataModule(
            train_dataset=self.train_dataset,
            val_dataset=self.val_dataset,
            test_dataset=self.test_dataset,
            train_config=self.train_config,
        )

        # Define Train Module
        self.train_module = TextClassifierModule(train_config=self.train_config)

        self.trainer = None

    def train(self, resume_from_checkpoint: Optional[str] = None):
        """
        Train the model, if resume_from_checkpoint is passed, resume from checkpoint.
        """
        # Resume from checkpoint, local path to a saved checkpoint.
        if resume_from_checkpoint:
            self.train_module = TextClassifierModule.load_from_checkpoint(
                resume_from_checkpoint
            )
        # Define Trainer and Callbacks for training, and start training
        self.base_train()


class FeedbackTextClassifier(BaseClassifier):
    """
    Feedback Text Classifier for training and inference. This class is used for training with additional feedback data.

    Args:
        train_dataset (Union[pd.DataFrame, List[List]]): Training dataset. Recommended to use the original training dataset.
        val_dataset (Union[pd.DataFrame, List[List]]): Validation dataset. Recommended to use the original validation dataset.
        feedback_dataset (Union[pd.DataFrame, List[List]]): Feedback dataset. It's okay even if the dataset consists of fewer than 10 items.
        test_dataset (Union[pd.DataFrame, List[List]], optional): Test dataset. Defaults to None.
        label_name_list (List, optional): List of label names. Defaults to None.
        train_config (TextClassifierConfig, optional): TextClassifierConfig. Defaults to None.
    """

    def __init__(
        self,
        train_dataset: Union[pd.DataFrame, List[List]],
        val_dataset: Union[pd.DataFrame, List[List]],
        feedback_dataset: Union[pd.DataFrame, List[List]],
        test_dataset: Union[pd.DataFrame, List[List]] = None,
        label_name_list: List = None,
        train_config: TextClassifierConfig = None,
    ):
        # Initial starting without loading from checkpoint
        if train_config is None:
            output_dir = "tmp_classifier"
            logger.info(
                f"No `TextClassifierConfig` passed, using `output_dir={output_dir}`."
            )
            train_config = TextClassifierConfig(output_dir=output_dir)

        else:
            if type(train_config) != TextClassifierConfig:
                raise ValueError(
                    f"train_config type has to be TextClassifierConfig, but got {type(train_config)}"
                )
        super().__init__(
            train_dataset,
            val_dataset,
            feedback_dataset,
            test_dataset,
            label_name_list,
            train_config,
        )

        # Define Data module
        self.data_module = FeedbackTextClassifierDataModule(
            train_dataset=self.train_dataset,
            val_dataset=self.val_dataset,
            test_dataset=self.test_dataset,
            train_config=self.train_config,
        )

        # Define Train Module
        self.train_module = FeedbackTextClassifierModule(train_config=self.train_config)

        self.trainer = None

    def train(self, resume_from_checkpoint=None):
        # Resume from checkpoint, local path to a saved checkpoint. Original checkpoint is required for feedback training.
        if resume_from_checkpoint is None:
            raise ValueError(
                "resume_from_checkpoint should be provided for feedback training."
            )

        self.train_module = FeedbackTextClassifierModule.load_from_checkpoint(
            resume_from_checkpoint, train_config=self.train_config
        )
        # Define Trainer and Callbacks for training, and start training
        self.base_train()



import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import torch

torch.backends.cudnn.benchmark = False
import logging
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import pytorch_lightning as pl

from ..trainer_util import set_seed
from .image_config import ImageClassifierConfig
from .image_datamodule import (
    ImageClassifierDataModule,
    FeedbackImageClassifierDataModule,
)
from .image_trainer import ImageClassifierModule, FeedbackImageClassifierModule
from ..base_classifier import BaseClassifier

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class ImageClassifier(BaseClassifier):
    """
    Image Classifier for training and inference.

    Args:
        train_dataset (Union[pd.DataFrame, List[List]]): Training dataset.
        val_dataset (Union[pd.DataFrame, List[List]]): Validation dataset.
        test_dataset (Union[pd.DataFrame, List[List]], optional): Test dataset. Defaults to None.
        label_name_list (List, optional): List of label names. Defaults to None.
        train_config (ImageClassifierConfig, optional): ImageClassifierConfig. Defaults to None.
    """

    def __init__(
        self,
        train_dataset: Union[pd.DataFrame, List[List]],
        val_dataset: Union[pd.DataFrame, List[List]],
        test_dataset: Union[pd.DataFrame, List[List]] = None,
        label_name_list: List = None,
        train_config: ImageClassifierConfig = None,
    ):
        # Initial starting without loading from checkpoint
        if train_config is None:
            output_dir = "tmp_classifier"
            logger.info(
                f"No `ImageClassifierConfig` passed, using `output_dir={output_dir}`."
            )
            train_config = ImageClassifierConfig(output_dir=output_dir)

        else:
            if type(train_config) != ImageClassifierConfig:
                raise ValueError(
                    f"train_config should be ImageClassifierConfig, but got {type(train_config)}"
                )

        feedback_dataset = None
        super().__init__(
            train_dataset,
            val_dataset,
            feedback_dataset,
            test_dataset,
            label_name_list,
            train_config,
        )

        # Define Data module
        self.data_module = ImageClassifierDataModule(
            train_dataset=self.train_dataset,
            val_dataset=self.val_dataset,
            test_dataset=self.test_dataset,
            train_config=self.train_config,
        )

        # Define Train Module
        self.train_module = ImageClassifierModule(train_config=self.train_config)

        self.trainer = None

    def train(self, resume_from_checkpoint: Optional[str] = None):
        # Resume from checkpoint, local path to a saved checkpoint.
        if resume_from_checkpoint:
            self.train_module = ImageClassifierModule.load_from_checkpoint(
                resume_from_checkpoint, train_config=self.train_config
            )
        # Define Trainer and Callbacks for training, and start training
        self.base_train()


class FeedbackImageClassifier(BaseClassifier):
    """
    Feedback Image Classifier for training and inference. This class is used for training with additional feedback data.

    Args:
        train_dataset (Union[pd.DataFrame, List[List]]): Training dataset. Recommended to use the original training dataset.
        val_dataset (Union[pd.DataFrame, List[List]]): Validation dataset. Recommended to use the original validation dataset.
        feedback_dataset (Union[pd.DataFrame, List[List]]): Feedback dataset. It's okay even if the dataset consists of fewer than 10 items.
        test_dataset (Union[pd.DataFrame, List[List]], optional): Test dataset. Defaults to None.
        label_name_list (List, optional): List of label names. Defaults to None.
        train_config (ImageClassifierConfig, optional): ImageClassifierConfig. Defaults to None.
    """

    def __init__(
        self,
        train_dataset: Union[pd.DataFrame, List[List]],
        val_dataset: Union[pd.DataFrame, List[List]],
        feedback_dataset: Union[pd.DataFrame, List[List]],
        test_dataset: Union[pd.DataFrame, List[List]] = None,
        label_name_list: List = None,
        train_config: ImageClassifierConfig = None,
    ):
        # Initial starting without loading from checkpoint
        if train_config is None:
            output_dir = "tmp_classifier"
            logger.info(
                f"No `ImageClassifierConfig` passed, using `output_dir={output_dir}`."
            )
            train_config = ImageClassifierConfig(output_dir=output_dir)

        else:
            if type(train_config) != ImageClassifierConfig:
                raise ValueError(
                    f"train_config should be ImageClassifierConfig, but got {type(train_config)}"
                )

        super().__init__(
            train_dataset,
            val_dataset,
            feedback_dataset,
            test_dataset,
            label_name_list,
            train_config,
        )

        # Define Data module
        self.data_module = FeedbackImageClassifierDataModule(
            train_dataset=self.train_dataset,
            val_dataset=self.val_dataset,
            test_dataset=self.test_dataset,
            train_config=self.train_config,
        )

        # Define Train Module
        self.train_module = FeedbackImageClassifierModule(
            train_config=self.train_config
        )

        self.trainer = None

    def train(self, resume_from_checkpoint=None):
        # Resume from checkpoint, local path to a saved checkpoint. Original checkpoint is required for feedback training.
        if resume_from_checkpoint is None:
            raise ValueError(
                "resume_from_checkpoint should be provided for feedback training."
            )

        self.train_module = FeedbackImageClassifierModule.load_from_checkpoint(
            resume_from_checkpoint, train_config=self.train_config
        )
        # Define Trainer and Callbacks for training, and start training
        self.base_train()



import torch
import os
from ..classifier import *
from typing import Union, List
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

from .util import postprocess


class PytorchClassificationPipeline:
    """
    Pytorch Classification Pipeline

    Args:
        model_type (str): Model type, support TextClassifier, ImageClassifier, ImageTextClassifier.
        save_dir (str): Model directory.
        device (str, optional): Device for inference, support cpu and gpu. Defaults to "gpu".
    """

    def __init__(
        self,
        model_type: str,
        save_dir: str,
        device: str = "gpu",
        quantization: bool = False,
        optimization: bool = False,
    ):
        # Device setting
        if device == "cpu":
            self.device = torch.device("cpu")
            self.autocast_device = "cpu"
        elif device == "gpu":
            self.device = torch.device("cuda")
            self.autocast_device = "cuda"
        else:
            raise ValueError(
                f"Not supported device type: {device}, choose in ['cpu', 'gpu']."
            )
        self.model_type = model_type
        # Use pytorch lightning module for inferene pipeline.
        if model_type.endswith("Module"):
            train_module_name = model_type
        else:
            train_module_name = model_type + "Module"
        # Call model class with model type string
        if train_module_name not in globals().keys():
            raise ValueError(f"Not supported model type: {model_type}")

        # Load pretrained model
        model_class = globals()[train_module_name]
        ckpt_path = os.path.join(save_dir, "model.ckpt")
        self.pretrained_model = model_class.load_from_checkpoint(
            ckpt_path, map_location=self.device
        )
        self.pretrained_model.eval()

        self.label_name_list = self.pretrained_model.train_config.label_name_list

    def __call__(
        self,
        inputs: Union[str, List[str]],
        topk: int = 1,
        return_logits: bool = False,
        sigmoid: bool = False,
        only_backbone: bool = False,
        last_embedding: bool = False,
        temperature: float = 1.0,
    ):
        """
        Args:
            inputs (Union[str, List[str]]): Input text or list of input texts.
            topk (int, optional): Top k prediction. Defaults to 1.
            return_logits (bool, optional): Return logits or not. Defaults to False.
            sigmoid (bool, optional): Apply sigmoid to logits or not. Defaults to False.
        """
        if only_backbone and last_embedding:
            raise ValueError(
                "only_backbone and last_embedding cannot be True at the same time."
            )

        if type(inputs) != list:
            inputs = [inputs]

        # In case when model is trained with multiple inputs, pipeline receives list of inputs. (e.g. ImageTextClassification)
        with torch.no_grad():
            if self.model_type == "ImageTextClassifier":
                if type(inputs[0]) != list:
                    inputs = list(map(lambda x: [x], inputs))

                if only_backbone:
                    backbone_output = self.pretrained_model.forward(
                        *inputs, eval=True, classifier=False
                    )
                    return backbone_output.float().detach().cpu().numpy()
                elif last_embedding:
                    if (
                        self.pretrained_model.classifier_head.__class__.__name__
                        != "Base2LinearClassifier"
                    ):
                        raise ValueError(
                            "last_embedding is only supported for Base2LinearClassifier."
                        )
                    backbone_output = self.pretrained_model.forward(
                        *inputs, eval=True, classifier=False
                    )
                    emb = self.pretrained_model.classifier_head.fc_layer1(
                        backbone_output.float()
                    )
                    emb = self.pretrained_model.classifier_head.bn1(emb)
                    emb = F.relu(emb)
                    return emb.detach().cpu().numpy()
                with torch.autocast(self.autocast_device):
                    logits = self.pretrained_model.forward(*inputs, eval=True)
            else:
                if self.model_type == "ImageClassifier":
                    if only_backbone:
                        backbone_output = self.pretrained_model.forward(
                            inputs, eval=True, classifier=False
                        )
                        return backbone_output.float().detach().cpu().numpy()
                    elif last_embedding:
                        if (
                            self.pretrained_model.classifier_head.__class__.__name__
                            != "Base2LinearClassifier"
                        ):
                            raise ValueError(
                                "last_embedding is only supported for Base2LinearClassifier."
                            )
                        backbone_output = self.pretrained_model.forward(
                            inputs, eval=True, classifier=False
                        )
                        emb = self.pretrained_model.classifier_head.fc_layer1(
                            backbone_output.float()
                        )
                        emb = self.pretrained_model.classifier_head.bn1(emb)
                        emb = F.relu(emb)
                        return emb.detach().cpu().numpy()
                    logits = self.pretrained_model.forward(inputs, eval=True)
                else:
                    if only_backbone:
                        backbone_output = self.pretrained_model.forward(
                            inputs, eval=True, classifier=False
                        )
                        return backbone_output.float().detach().cpu().numpy()
                    elif last_embedding:
                        if (
                            self.pretrained_model.classifier_head.__class__.__name__
                            != "Base2LinearClassifier"
                        ):
                            raise ValueError(
                                "last_embedding is only supported for Base2LinearClassifier."
                            )
                        backbone_output = self.pretrained_model.forward(
                            inputs, eval=True, classifier=False
                        )
                        emb = self.pretrained_model.classifier_head.fc_layer1(
                            backbone_output.float()
                        )
                        emb = self.pretrained_model.classifier_head.bn1(emb)
                        emb = F.relu(emb)
                        return emb.detach().cpu().numpy()
                    with torch.autocast(self.autocast_device):
                        logits = self.pretrained_model.forward(inputs, eval=True)

        return postprocess(
            logits,
            topk,
            self.label_name_list,
            return_logits,
            sigmoid,
            temperature,
        )

    def backbone_to_float32(self):
        self.pretrained_model = self.pretrained_model.float()



import os
import json
import pickle
from typing import Union, List
import torch
import torch.nn.functional as F

import logging
import numpy as np

logger = logging.getLogger(__name__)


def postprocess(
    logits: torch.Tensor,
    topk: int,
    label_name_list: List[str],
    return_logits: bool,
    sigmoid: bool,
    temperature: float,
):
    """
    Postprocess logits to topk predictions.

    Args:
        logits (torch.Tensor): Logits from model.
        topk (int): Top k prediction.
        label_name_list (List[str]): List of label names.
        return_logits (bool): Return logits or not.
        sigmoid (bool): Apply sigmoid to logits or not.
    """
    _, predictions = logits.topk(topk, 1, True, True)
    predictions_index = predictions.cpu().detach().numpy()

    cat_list = []
    # In case when label name list is not defined in the training step, pipeline returns the category label index of prediction, instead of category name.
    if label_name_list is None:
        logger.warning(
            "Label name list is not defined in the training step, pipeline returns the category label index of prediction, instead of category name."
        )

        for pred_index in predictions_index:
            cat_list.append(
                list(
                    map(
                        lambda x: x,
                        pred_index,
                    )
                )
            )
    else:
        for pred_index in predictions_index:
            cat_list.append(
                list(
                    map(
                        lambda x: label_name_list[x],
                        pred_index,
                    )
                )
            )

    if return_logits:
        if type(logits) == np.ndarray:
            logits = torch.from_numpy(logits)
        if sigmoid:
            return F.sigmoid(logits.to(torch.float32)).cpu().detach().numpy(), cat_list
        else:
            return (
                F.softmax(logits.to(torch.float32) / temperature, dim=-1)
                .cpu()
                .detach()
                .numpy(),
                cat_list,
            )

    else:
        return cat_list



import random
import numpy as np
import torch
import shutil
import os
import pickle
import torch
from typing import List, Union
from .custom_ops import custom_scaled_dot_product_attention
from PIL import Image, ImageOps
import requests
from optimum.onnxruntime import ORTModel
import json
import logging
from easydict import EasyDict
from transformers import AutoConfig
import torch.nn.functional as F  ## hm
from copy import deepcopy
import torch.nn as nn
import torchvision


logger = logging.getLogger(__name__)

from PIL import Image

if hasattr(Image, "Resampling"):
    _pil_interpolation_to_str = {
        Image.Resampling.NEAREST: "nearest",
        Image.Resampling.BILINEAR: "bilinear",
        Image.Resampling.BICUBIC: "bicubic",
        Image.Resampling.BOX: "box",
        Image.Resampling.HAMMING: "hamming",
        Image.Resampling.LANCZOS: "lanczos",
    }
else:
    _pil_interpolation_to_str = {
        Image.NEAREST: "nearest",
        Image.BILINEAR: "bilinear",
        Image.BICUBIC: "bicubic",
        Image.BOX: "box",
        Image.HAMMING: "hamming",
        Image.LANCZOS: "lanczos",
    }

_str_to_pil_interpolation = {v: k for k, v in _pil_interpolation_to_str.items()}


def sample_stratified(df, n_samples_per_label, problem_type):
    """
    Returns a stratified sample of rows from the dataframe, ensuring that each 'label' value is
    equally represented according to n_samples_per_label.

    Parameters:
    - df: DataFrame to sample from.
    - n_samples_per_label: Number of samples to draw for each label.

    Returns:
    - Stratified sample of the dataframe.
    """
    # Group by 'label' and sample n_samples_per_label from each group
    df_copy = df.copy()
    if problem_type == "multi_label_classification":
        # select random label in label_list
        df_copy["label"] = df_copy["label"].apply(lambda x: random.choice(x))
    sampled_df = df_copy.groupby("label", group_keys=False).apply(
        lambda x: x.sample(min(len(x), n_samples_per_label))
    )

    return sampled_df


def set_seed(seed: int):
    """
    Helper function for reproducible behavior to set the seed in `random`, `numpy`, `torch` and/or `tf` (if installed).

    Args:
        seed (`int`): The seed to set.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def model_save(model_type: str, save_dir: str, ckpt_path: str, device: str = "gpu"):
    """
    Save model backbone with huggingface format if possible, and classifier to onnx.
    Exporting backbone to onnx has to be done outside the trainer class, due to the large memory usage.
    Refer to the export_onnx function in trainer_util.py

    Args:
        model_type (str): model type, e.g. 'TextClassifier', 'ImageClassifier', 'ImageTextClassifier'
        save_dir (str): Directory to save the model
        ckpt_path (str): trained model checkpoint path
        device (str): device name, e.g. 'cpu', 'gpu'
    """
    if model_type == "TextClassifier":
        from .text_classifier import TextClassifierModule

        classifier_module = TextClassifierModule
    elif model_type == "ImageClassifier":
        from .image_classifier import ImageClassifierModule

        classifier_module = ImageClassifierModule
    elif model_type == "ImageTextClassifier":
        from .image_text_classifier import ImageTextClassifierModule

        classifier_module = ImageTextClassifierModule
    else:
        raise ValueError(f"Not supported model type: {model_type}")

    if device == "gpu":
        device = torch.device("cuda")
    elif device == "cpu":
        device = torch.device("cpu")
    else:
        raise ValueError(
            f"Not supported device type: {device}, choose in ['cpu', 'gpu']."
        )

    if model_type == "ImageTextClassifier":
        os.makedirs(os.path.join(save_dir, "image"), exist_ok=True)
        os.makedirs(os.path.join(save_dir, "text"), exist_ok=True)
    else:
        os.makedirs(save_dir, exist_ok=True)

    # Load pretrained model
    logger.info(f"Loading the checkpoint from {ckpt_path}")
    print(f"Loading the checkpoint from {ckpt_path}")
    train_module = classifier_module.load_from_checkpoint(
        ckpt_path, map_location=device
    )
    train_module.eval()

    # Contiguous for save_pretrained
    for param in train_module.parameters():
        param.data = param.data.contiguous()

    # For torch pipeline
    shutil.copy(ckpt_path, os.path.join(save_dir, "model.ckpt"))

    if model_type == "ImageTextClassifier":
        ### Image backbone to onnx
        image_backbone_save_module = train_module.image_backbone
        # Only for huggingface transformers
        if not train_module.train_config.image_timm:
            # TransformerWrapper
            image_backbone_save_module = image_backbone_save_module.model
            # When model is trained with LORA, we have to merge the lora weight with backbone.
            if train_module.train_config.image_lora:
                image_backbone_save_module = (
                    image_backbone_save_module.merge_and_unload()
                )
            image_backbone_save_module.save_pretrained(os.path.join(save_dir, "image"))
        ### Text backbone to onnx
        text_backbone_save_module = train_module.text_backbone
        # When model is trained with LORA, we have to merge the lora weight with backbone.
        if train_module.train_config.text_lora:
            text_backbone_save_module = text_backbone_save_module.merge_and_unload()
        # ONly for huggingface transformers
        assert (
            train_module.train_config.text_timm == False
        ), "Not supported timm model for text backbone."
        text_backbone_save_module.save_pretrained(os.path.join(save_dir, "text"))
        # Save Tokenizer
        train_module.tokenizer.save_pretrained(os.path.join(save_dir, "text"))
        train_module.text_initial_config.save_pretrained(os.path.join(save_dir, "text"))

        # with open(os.path.join(save_dir, "image", "config.json"), "w") as f:
        #     json.dump(train_module.image_initial_config, f)

        if train_module.image_initial_config is not None:
            train_module.image_initial_config.save_pretrained(save_dir)

        # Save validation transform with pickle
        with open(os.path.join(save_dir, "image", "val_transforms.pkl"), "wb") as f:
            pickle.dump(train_module.val_transforms, f)

    else:
        ### Backbone to onnx
        backbone_save_module = train_module.backbone
        # Only for huggingface transformers
        if not train_module.train_config.timm:
            # TransformerWrapper
            if model_type == "ImageClassifier":
                backbone_save_module = backbone_save_module.model
            # When model is trained with LORA, we have to merge the lora weight with backbone.
            if train_module.train_config.lora:
                backbone_save_module = backbone_save_module.merge_and_unload()
            backbone_save_module.save_pretrained(save_dir)

        if model_type == "TextClassifier":
            # Save Tokenizer
            train_module.tokenizer.save_pretrained(save_dir)
            train_module.initial_config.save_pretrained(save_dir)
        # ImageClassifier
        else:
            ### timm flag False
            if train_module.initial_config is not None:
                train_module.initial_config.save_pretrained(save_dir)
            # with open(os.path.join(save_dir, "config.json"), "w") as f:
            #     json.dump(train_module.initial_config, f)
            # Save validation transform with pickle
            with open(os.path.join(save_dir, "val_transforms.pkl"), "wb") as f:
                pickle.dump(train_module.val_transforms, f)

    # Save train config
    with open(os.path.join(save_dir, "train_config.json"), "w", encoding="utf-8") as f:
        json.dump(train_module.train_config.to_dict(), f, ensure_ascii=False, indent=4)

    ### Classifier save to onnx
    classifier_head_save_module = train_module.classifier_head
    dummy_input = torch.randn([1, train_module.embedding_size], device=device)
    torch.onnx.export(
        classifier_head_save_module,
        dummy_input,
        os.path.join(save_dir, "classifier_head.onnx"),
        input_names=["embeddings"],
        output_names=["logits"],
        dynamic_axes={"embeddings": {0: "batch_size"}, "logits": {0: "batch_size"}},
    )

    ## Classifier save to ckpt

    torch.save(
        classifier_head_save_module, os.path.join(save_dir, "classifier_head.ckpt")
    )


def export_onnx(model_type: str, save_dir: str, device: str = "gpu"):
    """
    Export Huggingface backbone model to onnx format.
    Args:
        model_type (`str`): Model type string, e.g. "TextClassifier", "ImageClassifier", "ImageTextClassifier"
        save_dir (`str`): Saved model directory, containing pytorch_model.bin or model.ckpt
        device (`str`): Device type for exporting onnx, no related with pipeline device  "cpu" or "gpu"
    """
    if device == "gpu":
        device = torch.device("cuda")
    elif device == "cpu":
        device = torch.device("cpu")
    else:
        raise ValueError(
            f"Not supported device type: {device}, choose in ['cpu', 'gpu']."
        )

    with open(os.path.join(save_dir, "train_config.json"), "r", encoding="utf-8") as f:
        train_config = EasyDict(json.load(f))

    if model_type == "TextClassifier":
        assert train_config.timm == False, "timm is not supported for TextClassifier"
        logger.info("Exporting torch pt model to onnx...")
        model = ORTModel.from_pretrained(save_dir, export=True)
        # Saved with model.onnx
        logger.info("Saving onnx file...")
        model.save_pretrained(save_dir)
    elif model_type == "ImageClassifier":
        from .image_classifier import ImageClassifierModule

        if train_config.timm:
            train_module = ImageClassifierModule.load_from_checkpoint(
                os.path.join(save_dir, "model.ckpt"), map_location=device
            )

            backbone_save_module = train_module.backbone
            from transformers import CLIPModel

            if isinstance(train_module.backbone, CLIPModel):
                backbone_save_module = backbone_save_module.vision_model

            if train_module.train_config.lora:
                backbone_save_module = backbone_save_module.merge_and_unload()

            torch.onnx.register_custom_op_symbolic(
                symbolic_name="aten::scaled_dot_product_attention",
                symbolic_fn=custom_scaled_dot_product_attention,
                opset_version=17,
            )

            # Custom operation registered, we can produce our onnx model

            dummy_input = torch.randn(
                train_module.data_config["input_size"], device=device
            ).unsqueeze(0)
            torch.onnx.export(
                backbone_save_module,
                dummy_input,
                os.path.join(save_dir, "model.onnx"),
                custom_opsets={"onnx-script": 17},
                opset_version=17,
                input_names=["inputs"],
                output_names=["embeddings"],
                dynamic_axes={
                    "inputs": {0: "batch_size"},
                    "embeddings": {0: "batch_size"},
                },
            )
        # Case for using huggingface transformers
        else:
            logger.info("Exporting torch pt model to onnx...")
            model = ORTModel.from_pretrained(save_dir, export=True)
            # Saved with model.onnx
            logger.info("Saving onnx file...")
            model.save_pretrained(save_dir)

    elif model_type == "ImageTextClassifier":
        from .image_text_classifier import ImageTextClassifierModule

        assert (
            train_config.text_timm == False
        ), "timm is not supported for TextClassifier"
        logger.info("Exporting torch pt text model to onnx...")
        text_model = ORTModel.from_pretrained(
            os.path.join(save_dir, "text"), export=True
        )
        # Saved with model.onnx
        logger.info("Saving onnx file...")
        text_model.save_pretrained(os.path.join(save_dir, "text"))

        logger.info("Exporting torch pt image model to onnx...")
        # huggingface transformers
        if not train_config.image_timm:
            image_model = ORTModel.from_pretrained(
                os.path.join(save_dir, "image"), export=True
            )
            # Saved with model.onnx
            logger.info("Saving onnx file...")
            image_model.save_pretrained(os.path.join(save_dir, "image"))
        # timm
        else:
            train_module = ImageTextClassifierModule.load_from_checkpoint(
                os.path.join(save_dir, "model.ckpt"), map_location=device
            )
            from transformers import CLIPModel

            if isinstance(train_module.image_backbone, CLIPModel):
                image_backbone_save_module = (
                    train_module.image_backbone.model.vision_model
                )
            else:
                image_backbone_save_module = train_module.image_backbone

            # image_backbone_save_module = train_module.image_backbone.model
            if train_module.train_config.image_lora:
                image_backbone_save_module = (
                    image_backbone_save_module.merge_and_unload()
                )

            torch.onnx.register_custom_op_symbolic(
                symbolic_name="aten::scaled_dot_product_attention",
                symbolic_fn=custom_scaled_dot_product_attention,
                opset_version=17,
            )
            # Custom operation registered, we can produce our onnx model

            dummy_image_input = torch.randn(
                train_module.image_data_config["input_size"], device=device
            ).unsqueeze(0)
            torch.onnx.export(
                image_backbone_save_module,
                dummy_image_input,
                os.path.join(save_dir, "image/model.onnx"),
                custom_opsets={"onnx-script": 17},
                opset_version=17,
                input_names=["inputs"],
                output_names=["embeddings"],
                dynamic_axes={
                    "inputs": {0: "batch_size"},
                    "embeddings": {0: "batch_size"},
                },
            )

    else:
        raise ValueError(f"Model type {model_type} not supported.")


def export_onnx_optimization(save_dir: str, optimization_level: int = 2):
    """
    export optimized onnx model using onnxruntime

    Args:
        save_dir (str): directory to save the optimized onnx model
        optimization_level (int, optional): optimization level. Defaults to 2. Can be 1, 2, 3, 4
    """
    from optimum.onnxruntime import ORTOptimizer, AutoOptimizationConfig

    optimization_config_dict = {
        1: AutoOptimizationConfig.O1(),
        2: AutoOptimizationConfig.O2(),
        3: AutoOptimizationConfig.O3(),
        4: AutoOptimizationConfig.O4(),
    }

    if optimization_level not in optimization_config_dict.keys():
        raise ValueError(f"Not supported optimization level {optimization_level}")
    optimization_config = optimization_config_dict[optimization_level]
    # Need to export onnx model first using export_onnx
    optimizer = ORTOptimizer.from_pretrained(save_dir, file_names="model.onnx")
    # saved with model_optimized.onnx
    optimizer.optimize(save_dir=save_dir, optimization_config=optimization_config)


def export_onnx_quantization(save_dir: str, cpu_architecture: str = "avx512"):
    """
    export quantized onnx model using onnxruntime

    Args:
        save_dir (str): directory to save the quantized onnx model
        cpu_architecture (str, optional): cpu architecture. Defaults to "avx512". Can be "arm64", "avx2", "avx512", "avx512_vnni", "tensorrt"
    """
    from optimum.onnxruntime import ORTQuantizer
    from optimum.onnxruntime.configuration import AutoQuantizationConfig

    supported_architectures = ["arm64", "avx2", "avx512", "avx512_vnni", "tensorrt"]
    if cpu_architecture not in supported_architectures:
        raise ValueError(
            f"Not supported quantization cpu architecture {cpu_architecture}"
        )
    # Need to export onnx model first using export_onnx
    quantizer = ORTQuantizer.from_pretrained(save_dir, file_name="model.onnx")
    # Use dynamic quantization, not using calibration dataset.
    dqconfig = getattr(AutoQuantizationConfig, cpu_architecture)(
        is_static=False, per_channel=False
    )
    # saved with model_quantized.onnx
    model_quantized_path = quantizer.quantize(
        save_dir=save_dir, quantization_config=dqconfig
    )


def ImageClassifierInputPreprocess(
    input_img_paths: List[str], device: torch.device, transforms
):
    """
    Preprocess input images for image classifier.

    Args:
        input_img_paths (List[str]): list of image paths
        device (torch.device): device to use
        transforms (torchvision.transforms): torchvision transforms
    """
    inputs = []
    # PIL open image then preprocess with augmentation.
    for img_path in input_img_paths:
        if type(img_path) == str:
            try:
                if img_path.startswith("http:") or img_path.startswith("https:"):
                    input_img = ImageOps.exif_transpose(
                        Image.open(requests.get(img_path, stream=True).raw)
                    ).convert("RGB")
                else:
                    input_img = ImageOps.exif_transpose(Image.open(img_path)).convert(
                        "RGB"
                    )
            except:
                logger.warning(
                    f"Cannot open image {img_path}, use zero padding image instead"
                )
                input_img = Image.fromarray(np.zeros((224, 224, 3), dtype=np.uint8))
        ### 보통 inference할 때만 사용. -> pil image 자체가 input인 경우.
        elif isinstance(img_path, Image.Image):
            input_img = img_path.convert("RGB")
        else:
            raise TypeError(
                "Not supported input type for image classifier. Only support img path or pil image."
            )
        input_tensor = transforms(input_img).to(device)
        inputs.append(input_tensor)

    return torch.stack(inputs)


def TextClassifierInputPreprocess(
    input_texts: List[str], text_prompt: str, device: torch.device, tokenizer
):
    """
    Preprocess input texts for text classifier, including prompt and tokenizer.

    Args:
        input_texts (List[str]): list of input texts
        text_prompt (str): text prompt
        device (torch.device): device to use
        tokenizer (transformers.tokenizer): tokenizer
    """
    # Prompting and apply tokenizer.
    prompted_input_texts = list(map(lambda x: text_prompt.format(x), input_texts))

    inputs = tokenizer(
        prompted_input_texts, padding=True, truncation=True, return_tensors="pt"
    ).to(device)

    return inputs


def TextClassifierSelectToken(
    backbone_outputs: torch.Tensor, inputs: dict, configs: AutoConfig
):
    """
    Selecting the proper token for text classifier.
    Can be different for each model_type, so we need to handle it separately.
    Currently only implemented for gpt_neox.

    Args:
        backbone_outputs (torch.tensor): backbone outputs
        inputs (dict): model inputs
        configs (AutoConfig): model configs
    """
    # For classificaiton, model use the last token as other causal models (e.g. GPT-1) do. We utilize attention mask to know the position of the last token.
    if configs.model_type == "gpt_neox":
        last_hidden_state = backbone_outputs[0]
        if type(last_hidden_state) == np.ndarray:
            selected_tokens = last_hidden_state[
                np.arange(last_hidden_state.shape[0]),
                inputs["attention_mask"].sum(axis=-1) - 1,
            ]
            selected_tokens = torch.from_numpy(selected_tokens)
        else:
            selected_tokens = last_hidden_state[
                torch.arange(last_hidden_state.shape[0]),
                inputs["attention_mask"].sum(axis=-1) - 1,
            ]

        return selected_tokens

    elif (
        configs.model_type == "roberta"
        or configs.model_type == "xlm-roberta"
        or configs.model_type == "electra"
        or configs.model_type == "bert"
    ):
        last_hidden_state = backbone_outputs[0]
        if type(last_hidden_state) == np.ndarray:
            last_hidden_state = torch.from_numpy(last_hidden_state)
        selected_tokens = last_hidden_state[:, 0]

        return selected_tokens
    ##  hm
    elif configs.model_type == "mpnet":
        # print('============= you are here =============')
        def mean_pooling(model_output, attention_mask):
            token_embeddings = model_output[0]
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
                input_mask_expanded.sum(1), min=1e-9
            )

        attention_mask = inputs["attention_mask"]

        if type(backbone_outputs) == np.ndarray:
            backbone_outputs = torch.from_numpy(backbone_outputs)

        selected_tokens = mean_pooling(backbone_outputs, attention_mask)
        selected_tokens = F.normalize(selected_tokens, p=2, dim=1)

        return selected_tokens

    else:
        raise NotImplementedError(
            f"{configs.model_type} selecting token method is not implemented."
        )


def is_parallel(model):
    """check if model is in parallel mode."""

    parallel_type = (
        nn.parallel.DataParallel,
        nn.parallel.DistributedDataParallel,
    )
    return isinstance(model, parallel_type)


def copy_attr(a, b, include=(), exclude=()):
    # Copy attributes from b to a, options to only include [...] and to exclude [...]
    for k, v in b.__dict__.items():
        if (len(include) and k not in include) or k.startswith("_") or k in exclude:
            continue
        else:
            setattr(a, k, v)


class ModelEMA:
    """
    Model Exponential Moving Average from https://github.com/rwightman/pytorch-image-models
    Keep a moving average of everything in the model state_dict (parameters and buffers).
    This is intended to allow functionality like
    https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    A smoothed version of the weights is necessary for some training schemes to perform well.
    This class is sensitive where it is initialized in the sequence of model init,
    GPU assignment and distributed training wrappers.
    """

    def __init__(self, models, decay=0.9999, updates=0):
        """
        Args:
            models (tuple of nn.Module): model list to apply EMA.
            decay (float): ema decay reate.
            updates (int): counter of EMA updates.
        """
        # Create EMA(FP32)
        self.ema = ()
        self.updates = ()
        self.decay = decay
        for model in models:
            ema = deepcopy(model.module if is_parallel(model) else model).eval()
            self.ema = self.ema + (ema,)
            # decay exponential ramp (to help early epochs)
            # self.decay = lambda x: decay * (1 - math.exp(-x / 2000))
            for p in ema.parameters():
                p.requires_grad_(False)

        self.updates = (updates,) * len(models)

    def update(self, models):
        # Update EMA parameters
        self.updates += 1
        for idx, model in enumerate(models):
            with torch.no_grad():
                # d = self.decay(self.updates)
                d = self.decay
                msd = (
                    model.module.state_dict()
                    if is_parallel(model)
                    else model.state_dict()
                )  # model state_dict
                for k, v in self.ema[idx].state_dict().items():
                    if v.dtype.is_floating_point:
                        v *= d
                        v += (1.0 - d) * msd[k].detach()

    @staticmethod
    def update_attr(
        updating_models,
        cand_models,
        include=(),
        exclude=("process_group", "reducer", "training"),
    ):
        for updating_model, cand_model in zip(updating_models, cand_models):
            # Update EMA attributes
            copy_attr(updating_model, cand_model, include, exclude)


def find_indices_based_on_max_and_label(tensor, labels, threshold):
    # 각 행의 최대값과 그 인덱스를 계산
    max_values, max_indices = torch.max(tensor, dim=1)

    # 조건 1: 최대값이 threshold보다 큰 경우
    condition1 = max_values > threshold

    # 조건 2: 최대값의 인덱스가 labels와 일치하는 경우
    condition2 = max_indices == labels

    # 두 조건을 모두 만족하는 경우
    valid_indices = torch.where(condition1 & condition2)[0]

    # 두 조건 중 하나라도 만족하지 않는 경우
    invalid_indices = torch.where(~(condition1 & condition2))[0]

    return valid_indices, invalid_indices


class SquarePad:
    def __call__(self, image):
        w, h = image.size
        max_wh = np.max([w, h])
        hp = int((max_wh - w) / 2)
        vp = int((max_wh - h) / 2)
        padding = (hp, vp, hp, vp)
        return torchvision.transforms.functional.pad(image, padding, 0, "constant")


class TransformerWrapper(nn.Module):
    """
    Wrapper for transformers model to use in PyTorch Lightning (compatible with timm)
    """

    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, x):
        return self.model(pixel_values=x).pooler_output



import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"
import copy
import torch

torch.backends.cudnn.benchmark = False
import logging
from typing import List, Union

import pandas as pd
from PIL import Image
import requests

# for test
from tqdm import tqdm
import time
from .zeroshot_image_datamodule import TextDataset, ImageDataset, ImageClassifierDataset
from .zeroshot_image_config import ZeroshotImageClassifierConfig
import torch.nn.functional as F
from torch.utils.data import DataLoader

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class ZeroshotImageClassifier:
    """
    Zeroshot Image Classifier.

    Args:
        test_dataset (Union[pd.DataFrame, List[List]]): Test dataset, optional.
        label_name_list (List[str]): List of label names, optional.
        label_embeddings (torch.Tensor): Label embeddings, optional.
        config (ZeroshotImageClassifierConfig): ZeroshotImageClassifierConfig, optional.
        device (str): Device, optional.
    """

    def __init__(
        self,
        test_dataset: Union[pd.DataFrame, List[List]] = None,
        label_name_list: List[str] = None,
        label_embeddings: torch.Tensor = None,
        config: ZeroshotImageClassifierConfig = None,
        device: str = None,
    ):
        # Initial starting without loading from checkpoint
        if config is None:
            output_dir = "tmp_classifier"
            logger.info(
                f"No `ZeroshotImageClassifierConfig` passed, using `output_dir={output_dir}`."
            )
            train_config = ZeroshotImageClassifierConfig(output_dir=output_dir)

        self.train_config = config

        if device is None:
            self.device = self.train_config.device
        else:
            self.device = device

        test_dataset = copy.deepcopy(test_dataset)
        # Dataset Setting

        if test_dataset is not None:
            logger.info(
                "Test dataset is provided, you can inference test dataset with ZeroshotImageClassifier().test()"
            )
            if type(test_dataset) == pd.DataFrame:
                if len(test_dataset.columns) != 2:
                    raise ValueError(
                        f"When the test dataset data type is DataFrame, length of the column has to fixed with 2."
                    )
                if test_dataset.columns[0] != "img_path":
                    logger.warning(
                        f"When the test dataset data type is DataFrame, first column name has to be 'img_path', assuming '{test_dataset.columns[0]}' means 'img_path' column."
                    )
                    test_dataset.rename(
                        columns={test_dataset.columns[0]: "img_path"}, inplace=True
                    )
                if test_dataset.columns[1] != "label":
                    logger.warning(
                        f"When the test dataset data type is DataFrame, first column name has to be 'label', assuming '{test_dataset.columns[1]}' means 'label' column."
                    )
                    test_dataset.rename(
                        columns={test_dataset.columns[1]: "label"}, inplace=True
                    )

            elif type(test_dataset) == list:
                if len(test_dataset) != 2:
                    raise ValueError(
                        f"When the test dataset data type is List, length of the list has to fixed with 2."
                    )
                else:
                    logger.warning(
                        "When the test dataset data type is List, we assume that the first component of list means 'img_path', and the second component of list means 'label'."
                    )
                # test_dataset = pd.DataFrame(test_dataset, columns=["img_path", "label"])
                test_dataset = pd.DataFrame(
                    {"img_path": test_dataset[0], "label": test_dataset[1]}
                )

            else:
                raise TypeError(
                    f"Not supported test dataset type, {type(test_dataset)}"
                )

            if test_dataset["label"].dtype != "int":
                raise TypeError(
                    f"Expect int datatype for test dataset label, but get {test_dataset['label'].dtype}"
                )

        if self.train_config.backbone_name == "clip":
            from transformers import AutoModel, AutoProcessor, AutoTokenizer

            name = self.train_config.model_name
            self.model = AutoModel.from_pretrained(name).to(self.device)
            self.processor = AutoProcessor.from_pretrained(name)
            self.tokenizer = AutoTokenizer.from_pretrained(name)

        else:
            raise ValueError(
                "This backbone is not currently supported. Supported backbone list are ['clip']."
            )

        if label_embeddings is None:
            if label_name_list is None:
                raise ValueError(
                    f"You must enter either 'label_name_list' or 'label_embeddings'."
                )
            else:
                logger.info("Make label embeddings with label name list.")
                dataset = TextDataset(label_name_list)
                dataloader = DataLoader(dataset=dataset, batch_size=1)
                text_embeddings = []
                success_text = []
                with torch.no_grad():
                    for batch in dataloader:
                        try:
                            inputs = self.tokenizer(
                                batch, padding=True, return_tensors="pt"
                            ).to(self.device)
                            emb = self.model.get_text_features(**inputs)
                            text_embeddings.append(emb)
                            success_text.append(batch)
                        except:
                            logger.warning(f"Cannot tokenize text, {batch}")
                    label_embeddings = torch.cat(text_embeddings).to(self.device)

        self.label_name_list = success_text
        self.label_embeddings = label_embeddings.to(self.device)

        # Check number of class when training config num class value is default value.
        if label_name_list is not None:
            if label_embeddings is not None:
                if label_embeddings.size()[0] != len(label_name_list):
                    raise ValueError(
                        f"The length of the label_name_list and size of the label_embeddings are different. Please check the size."
                    )

            logger.info("Get number of class with the length of label name list.")
            self.train_config.num_class = len(label_name_list)
            self.train_config.label_name_list = label_name_list
        else:
            if label_embeddings is not None:
                logger.info(
                    "Get number of class with the length of label embeddings tensor."
                )
                self.train_config.num_class = label_embeddings.size[0]
            else:
                raise ValueError(
                    f"You must enter either 'label_name_list' or 'label_embeddings'."
                )

        if test_dataset is not None:
            self.test_dataset = test_dataset

    def make_embedding(self, text: List[str] = None, image_path: List[str] = None):
        """
        Make embedding with text or image.

        Args:
            text (List[str], optional): List of text. Defaults to None.
            image_path (List[str], optional): List of image path. Defaults to None.
        """
        if text is None and image_path is None:
            raise ValueError("You must enter text or image_path.")
        if text is not None and image_path is not None:
            raise ValueError("Please enter either an text or image_path.")

        if image_path:
            dataset = ImageDataset(image_path)
            dataloader = DataLoader(dataset=dataset, batch_size=1)
            image_embeddings = []
            process_data = []
            with torch.no_grad():
                for batch in dataloader:
                    path = batch[0]
                    try:
                        if path[:4] == "http":
                            image = Image.open(requests.get(path, stream=True).raw)
                        else:
                            image = Image.open(path)

                        inputs = self.processor(images=image, return_tensors="pt").to(
                            self.device
                        )
                        emb = self.model.get_image_features(**inputs)
                        image_embeddings.append(emb)
                        process_data.extend(batch)
                    except:
                        logger.warning(f"Cannot processing image, {path}")
                result = torch.cat(image_embeddings).to(self.device)

        if text:
            dataset = TextDataset(text)
            dataloader = DataLoader(dataset=dataset, batch_size=1)
            text_embeddings = []
            process_data = []
            with torch.no_grad():
                for batch in dataloader:
                    try:
                        inputs = self.tokenizer(
                            batch, padding=True, return_tensors="pt"
                        ).to(self.device)
                        emb = self.model.get_text_features(**inputs)
                        text_embeddings.append(emb)
                        process_data.extend(batch)
                    except:
                        logger.warning(f"Cannot tokenize text, '{batch[0]}'")
                result = torch.cat(text_embeddings).to(self.device)

        return result, process_data

    def top_cos_sim(self, query, corpus, k):
        cos_sim = F.normalize(query) @ F.normalize(corpus).T
        similarity, indices = torch.topk(cos_sim, k)
        return similarity.cpu(), indices.cpu()

    def test(self):
        """
        Test model with test dataset.
        """
        print("Start testing")
        start_time = time.time()
        topk_list = self.train_config.print_topk_list
        if max(topk_list) > self.train_config.num_class:
            logger.warning(f"max of the topk is bigger than the number of class")
            topk_list = [i for i in topk_list if i <= self.train_config.num_class]

        test_dataset = ImageClassifierDataset(self.test_dataset)
        dataloader = DataLoader(dataset=test_dataset, batch_size=1)

        process_data = []

        total_num = 0
        topk_count = [0] * len(topk_list)
        # pred = []

        with torch.no_grad():
            for img_path, label in tqdm(dataloader):
                path = img_path[0]
                label = label[0].item()

                try:
                    if path[:4] == "http":
                        image = Image.open(requests.get(path, stream=True).raw)
                    else:
                        image = Image.open(path)

                    inputs = self.processor(images=image, return_tensors="pt").to(
                        self.device
                    )
                    emb = self.model.get_image_features(**inputs)
                    process_data.extend(img_path)

                    cos_sim, cos_ids = self.top_cos_sim(
                        emb, self.label_embeddings, max(topk_list)
                    )
                    # pred.append(cos_ids[0].tolist())

                    total_num += 1
                    for topk_idx, topk in enumerate(topk_list):
                        if label in cos_ids[0][:topk]:
                            topk_count[topk_idx] += 1

                except:
                    logger.warning(f"Cannot processing image, {path}")

        topk_acc = [round(i / total_num, 4) for i in topk_count]
        for acc, topk in zip(topk_acc, topk_list):
            print(f"Top {topk} accuracy: {acc}")
            logger.info(f"Top {topk} accuracy: {acc}")

        end_time = time.time()
        print(f"Total testing time: {end_time - start_time} seconds")

    def inference(self, input_image_path, topk=5):
        print("Start testing")
        start_time = time.time()
        if topk > self.train_config.num_class:
            topk = self.train_config.num_class
            logger.warning(
                f"topk is bigger than the number of class. topk is set to the number of classes '{topk}'"
            )

        test_dataset = ImageClassifierDataset(self.test_dataset)
        dataloader = DataLoader(dataset=test_dataset, batch_size=1)

        process_image_path = []
        pred = []

        with torch.no_grad():
            for path in input_image_path:
                try:
                    if path[:4] == "http":
                        image = Image.open(requests.get(path, stream=True).raw)
                    else:
                        image = Image.open(path)

                    inputs = self.processor(images=image, return_tensors="pt").to(
                        self.device
                    )
                    emb = self.model.get_image_features(**inputs)
                    process_image_path.append(path)

                    cos_sim, cos_ids = self.top_cos_sim(
                        emb, self.label_embeddings, topk
                    )

                    pred_index = cos_ids[0].tolist()
                    pred_category = [self.label_name_list[i] for i in pred_index]

                    pred.append(pred_category)

                except:
                    logger.warning(f"Cannot processing image, {path}")

        end_time = time.time()
        print(f"Total inference number: {len(process_image_path)} images")
        print(f"Total inference time: {end_time - start_time} seconds")
        return process_image_path, pred

    """
    Save model backbone with huggingface format, and classifier to onnx. 
    Exporting backbone to onnx has to be done outside the trainer class, due to the large memory usage. 
    Refer to the export_onnx function in trainer_util.py 
    """



import onnxscript
import torch

# Assuming you use opset18
from onnxscript.onnx_opset import opset17 as op

custom_opset = onnxscript.values.Opset(domain="onnx-script", version=17)


@onnxscript.script(custom_opset)
def ScaledDotProductAttention(
    query,
    key,
    value,
    dropout_p,
):
    # Swap the last two axes of key
    key_shape = op.Shape(key)
    key_last_dim = key_shape[-1:]
    key_second_last_dim = key_shape[-2:-1]
    key_first_dims = key_shape[:-2]
    # Contract the dimensions that are not the last two so we can transpose
    # with a static permutation.
    key_squeezed_shape = op.Concat(
        op.Constant(value_ints=[-1]), key_second_last_dim, key_last_dim, axis=0
    )
    key_squeezed = op.Reshape(key, key_squeezed_shape)
    key_squeezed_transposed = op.Transpose(key_squeezed, perm=[0, 2, 1])
    key_transposed_shape = op.Concat(
        key_first_dims, key_last_dim, key_second_last_dim, axis=0
    )
    key_transposed = op.Reshape(key_squeezed_transposed, key_transposed_shape)

    embedding_size = op.CastLike(op.Shape(query)[-1], query)
    scale = op.Div(1.0, op.Sqrt(embedding_size))

    # Scale q, k before matmul for stability see https://tinyurl.com/sudb9s96 for math
    query_scaled = op.Mul(query, op.Sqrt(scale))
    key_transposed_scaled = op.Mul(key_transposed, op.Sqrt(scale))
    attn_weight = op.Softmax(
        op.MatMul(query_scaled, key_transposed_scaled),
        axis=-1,
    )
    attn_weight, _ = op.Dropout(attn_weight, dropout_p)
    return op.MatMul(attn_weight, value)


# ScaledDotProductAttention.opname = ScaledDotProductAttention.name


def custom_scaled_dot_product_attention(
    g, query, key, value, attn_mask, dropout, is_causal, scale=None
):
    return g.onnxscript_op(
        ScaledDotProductAttention, query, key, value, dropout
    ).setType(query.type())



import os
import json
import pickle
from typing import Union, List

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
import onnxruntime
from ..classifier.trainer_util import (
    TextClassifierInputPreprocess,
    TextClassifierSelectToken,
    ImageClassifierInputPreprocess,
)
from transformers import AutoTokenizer, AutoConfig
import torch
import torch.nn.functional as F
from easydict import EasyDict

from .onnx_util import onnx_inference, classifier_onnx_preprocess
from .util import postprocess


class OnnxTextClassificationPipeline:
    """
    Onnx Text Classification Pipeline

    Args:
        model_type (str): Model type, only support TextClassifier.
        save_dir (str): Model directory.
        device (str, optional): Device for inference, support cpu and gpu. Defaults to "cpu".
    """

    def __init__(
        self,
        model_type: str,
        save_dir: str,
        device: str = "cpu",
        quantization: bool = False,
        optimization: bool = False,
    ):
        self.tokenizer = AutoTokenizer.from_pretrained(save_dir)

        if device == "gpu":
            self.device = torch.device("cuda")
            if quantization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_quantized.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            elif optimization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_optimized.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            else:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            self.classifier_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "classifier_head.onnx"),
                providers=["CUDAExecutionProvider"],
            )
        # cpu mode
        else:
            self.device = torch.device("cpu")
            if quantization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_quantized.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            elif optimization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_optimized.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            else:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            self.classifier_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "classifier_head.onnx"),
                providers=["CPUExecutionProvider"],
            )

        # Load train config
        with open(
            os.path.join(save_dir, "train_config.json"), "r", encoding="utf-8"
        ) as f:
            train_config = EasyDict(json.load(f))
        self.text_prompt = train_config.text_prompt
        self.label_name_list = train_config.label_name_list
        assert (
            train_config.timm == False
        ), "timm model is not supported yet for TextClassificationPipeline"
        self.config = AutoConfig.from_pretrained(save_dir)

        self.classifier_head = torch.load(
            os.path.join(save_dir, "classifier_head.ckpt"), map_location=self.device
        )
        self.classifier_head.eval()

    def __call__(
        self,
        inputs: Union[str, List[str]],
        topk: int = 1,
        return_logits: bool = False,
        sigmoid: bool = False,
        only_backbone: bool = False,
        last_embedding: bool = False,
        temperature: float = 1.0,
    ):
        """
        Args:
            inputs (Union[str, List[str]]): Input text or list of input text.
            topk (int, optional): Top k prediction. Defaults to 1.
            return_logits (bool, optional): Return logits or not. Defaults to False.
            sigmoid (bool, optional): Apply sigmoid or not(Softmax). Defaults to False.
            only_backbone (bool, optional): Only return backbone output embedding. Defaults to False.
        """
        if only_backbone and last_embedding:
            raise ValueError(
                "only_backbone and last_embedding cannot be True at the same time."
            )

        if type(inputs) != list:
            inputs = [inputs]

        preprocessed_inputs = TextClassifierInputPreprocess(
            inputs,
            text_prompt=self.text_prompt,
            device=self.device,
            tokenizer=self.tokenizer,
        )

        embeddings = onnx_inference(self.ort_session, preprocessed_inputs, self.device)

        selected_token_embedding = TextClassifierSelectToken(
            [embeddings], preprocessed_inputs, self.config
        )

        if only_backbone:
            return selected_token_embedding.cpu().detach().numpy()

        elif last_embedding:
            if self.classifier_head.__class__.__name__ != "Base2LinearClassifier":
                raise ValueError(
                    "last_embedding is only supported for Base2LinearClassifier."
                )
            emb = self.classifier_head.fc_layer1(
                selected_token_embedding.to(self.device)
            )
            emb = self.classifier_head.bn1(emb)
            emb = F.relu(emb)
            return emb.detach().cpu().numpy()

        # CLassifier ort input preprocess
        classifier_ort_input_dict = classifier_onnx_preprocess(
            self.classifier_ort_session, selected_token_embedding
        )

        logits = onnx_inference(
            self.classifier_ort_session, classifier_ort_input_dict, self.device
        )

        return postprocess(
            logits, topk, self.label_name_list, return_logits, sigmoid, temperature
        )


class OnnxImageClassificationPipeline:
    """
    Onnx Image Classification Pipeline

    Args:
        model_type (str): Model type, only support ImageClassifier.
        save_dir (str): Model directory.
        device (str, optional): Device for inference, support cpu and gpu. Defaults to "cpu".
    """

    def __init__(
        self,
        model_type: str,
        save_dir: str,
        device: str = "cpu",
        quantization: bool = False,
        optimization: bool = False,
    ):
        if device == "gpu":
            self.device = torch.device("cuda")
            if quantization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_quantized.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            elif optimization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_optimized.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            else:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            self.classifier_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "classifier_head.onnx"),
                providers=["CUDAExecutionProvider"],
            )
        # cpu mode
        else:
            self.device = torch.device("cpu")
            if quantization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_quantized.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            elif optimization:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model_optimized.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            else:
                self.ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "model.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            self.classifier_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "classifier_head.onnx"),
                providers=["CPUExecutionProvider"],
            )
        # Load train config
        with open(
            os.path.join(save_dir, "train_config.json"), "r", encoding="utf-8"
        ) as f:
            train_config = EasyDict(json.load(f))
        self.label_name_list = train_config.label_name_list

        # Load val_transforms
        with open(os.path.join(save_dir, "val_transforms.pkl"), "rb") as f:
            self.transforms = pickle.load(f)

        # Load config
        if os.path.exists(os.path.join(save_dir, "config.json")):
            with open(os.path.join(save_dir, "config.json"), "r") as f:
                self.config = json.load(f)
        else:
            self.config = None

        self.classifier_head = torch.load(
            os.path.join(save_dir, "classifier_head.ckpt"), map_location=self.device
        )
        self.classifier_head.eval()

    def __call__(
        self,
        inputs: Union[str, List[str]],
        topk: int = 1,
        return_logits: bool = False,
        sigmoid: bool = False,
        only_backbone: bool = False,
        last_embedding: bool = False,
        temperature: float = 1.0,
    ):
        """
        Args:
            inputs (Union[str, List[str]]): Input image path or list of input image path.
            topk (int, optional): Top k prediction. Defaults to 1.
            return_logits (bool, optional): Return logits or not. Defaults to False.
            sigmoid (bool, optional): Apply sigmoid or not(Softmax). Defaults to False.
            only_backbone (bool, optional): Only return backbone output embedding. Defaults to False.
        """
        if only_backbone and last_embedding:
            return ValueError(
                "only_backbone and last_embedding cannot be True at the same time."
            )

        if type(inputs) != list:
            inputs = [inputs]

        #  preprocessed_inputs = TextClassifierInputPreprocess(
        #     inputs,
        #     text_prompt=self.text_prompt,
        #     device=self.device,
        #     tokenizer=self.tokenizer,
        # )

        # embeddings = onnx_inference(self.ort_session, preprocessed_inputs, self.device)

        preprocessed_inputs = {
            self.ort_session.get_inputs()[0].name: ImageClassifierInputPreprocess(
                inputs,
                device=self.device,
                transforms=self.transforms,
            )
        }

        embeddings = onnx_inference(self.ort_session, preprocessed_inputs, self.device)

        ### Transformer 사용시에는 select token 과정 필요.
        if self.config:
            if self.config["model_type"] == "vit":
                embeddings = embeddings[:, 0]
            else:
                raise NotImplementedError("Only vit is supported for now")

        if only_backbone:
            return embeddings.cpu().detach().numpy()

        elif last_embedding:
            if self.classifier_head.__class__.__name__ != "Base2LinearClassifier":
                raise ValueError(
                    "last_embedding is only supported for Base2LinearClassifier."
                )
            emb = self.classifier_head.fc_layer1(embeddings.to(self.device))
            emb = self.classifier_head.bn1(emb)
            emb = F.relu(emb)
            return emb.detach().cpu().numpy()

        # CLassifier ort input preprocess
        classifier_ort_input_dict = classifier_onnx_preprocess(
            self.classifier_ort_session, embeddings
        )
        logits = onnx_inference(
            self.classifier_ort_session, classifier_ort_input_dict, self.device
        )

        return postprocess(
            logits, topk, self.label_name_list, return_logits, sigmoid, temperature
        )


class OnnxImageTextClassificationPipeline:
    """
    Onnx Image Text Classification Pipeline

    Args:
        model_type (str): Model type, only support ImageTextClassifier.
        save_dir (str): Model directory.
        device (str, optional): Device for inference, support cpu and gpu. Defaults to "cpu".
    """

    def __init__(
        self,
        model_type: str,
        save_dir: str,
        device: str = "cpu",
        quantization: bool = False,
        optimization: bool = False,
    ):
        self.tokenizer = AutoTokenizer.from_pretrained(os.path.join(save_dir, "text"))

        if device == "gpu":
            self.device = torch.device("cuda")
            self.image_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "image/model.onnx"),
                providers=["CUDAExecutionProvider"],
            )
            if quantization:
                self.text_ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "text/model_quantized.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            elif optimization:
                self.text_ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "text/model_optimized.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            else:
                self.text_ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "text/model.onnx"),
                    providers=["CUDAExecutionProvider"],
                )
            self.classifier_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "classifier_head.onnx"),
                providers=["CUDAExecutionProvider"],
            )
        # CPU mode
        else:
            self.device = torch.device("cpu")
            self.image_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "image/model.onnx"),
                providers=["CPUExecutionProvider"],
            )
            if quantization:
                self.text_ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "text/model_quantized.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            elif optimization:
                self.text_ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "text/model_optimized.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            else:
                self.text_ort_session = onnxruntime.InferenceSession(
                    os.path.join(save_dir, "text/model.onnx"),
                    providers=["CPUExecutionProvider"],
                )
            self.classifier_ort_session = onnxruntime.InferenceSession(
                os.path.join(save_dir, "classifier_head.onnx"),
                providers=["CPUExecutionProvider"],
            )
        # Load Train config
        with open(
            os.path.join(save_dir, "train_config.json"), "r", encoding="utf-8"
        ) as f:
            train_config = EasyDict(json.load(f))
        self.text_prompt = train_config.text_prompt
        self.label_name_list = train_config.label_name_list
        self.text_config = AutoConfig.from_pretrained(os.path.join(save_dir, "text"))

        # Load val_transforms
        with open(os.path.join(save_dir, "image", "val_transforms.pkl"), "rb") as f:
            self.transforms = pickle.load(f)

        self.classifier_head = torch.load(
            os.path.join(save_dir, "classifier_head.ckpt"), map_location=self.device
        )
        self.classifier_head.eval()

    def __call__(
        self,
        image_inputs: Union[str, List[str]],
        text_inputs: Union[str, List[str]],
        topk: int = 1,
        return_logits: bool = False,
        sigmoid: bool = False,
        only_backbone: bool = False,
        last_embedding: bool = False,
        temperature: float = 1.0,
    ):
        """
        Args:
            image_inputs (Union[str, List[str]]): Input image path or list of input image path.
            text_inputs (Union[str, List[str]]): Input text or list of input text.
            topk (int, optional): Top k prediction. Defaults to 1.
            return_logits (bool, optional): Return logits or not. Defaults to False.
            sigmoid (bool, optional): Apply sigmoid or not(Softmax). Defaults to False.
            only_backbone (bool, optional): Only return backbone output. Defaults to False.
        """
        if only_backbone and last_embedding:
            raise ValueError(
                "only_backbone and last_embedding cannot be True at the same time."
            )

        if type(image_inputs) == str:
            image_inputs = [image_inputs]

        if type(text_inputs) == str:
            text_inputs = [text_inputs]

        preprocessed_image_inputs = {
            "inputs": ImageClassifierInputPreprocess(
                image_inputs,
                device=self.device,
                transforms=self.transforms,
            )
        }

        preprocessed_text_inputs = TextClassifierInputPreprocess(
            text_inputs,
            text_prompt=self.text_prompt,
            device=self.device,
            tokenizer=self.tokenizer,
        )

        image_embeddings = onnx_inference(
            self.image_ort_session, preprocessed_image_inputs, self.device
        )
        text_embeddings = onnx_inference(
            self.text_ort_session, preprocessed_text_inputs, self.device
        )

        selected_token_embedding = TextClassifierSelectToken(
            [text_embeddings], preprocessed_text_inputs, self.text_config
        )

        embeddings = torch.cat([image_embeddings, selected_token_embedding], dim=-1)

        if only_backbone:
            return embeddings.cpu().detach().numpy()

        elif last_embedding:
            if self.classifier_head.__class__.__name__ != "Base2LinearClassifier":
                raise ValueError(
                    "last_embedding is only supported for Base2LinearClassifier."
                )
            emb = self.classifier_head.fc_layer1(embeddings.to(self.device))
            emb = self.classifier_head.bn1(emb)
            emb = F.relu(emb)
            return emb.detach().cpu().numpy()

        # CLassifier ort input preprocess
        classifier_ort_input_dict = classifier_onnx_preprocess(
            self.classifier_ort_session, embeddings
        )

        logits = onnx_inference(
            self.classifier_ort_session, classifier_ort_input_dict, self.device
        )

        return postprocess(
            logits, topk, self.label_name_list, return_logits, sigmoid, temperature
        )



from .image_text_classifier import ImageTextClassifier, FeedbackImageTextClassifier
from .image_text_config import ImageTextClassifierConfig
from .image_text_trainer import ImageTextClassifierModule, FeedbackImageTextClassifierModule



from pytorch_lightning import LightningDataModule
import pandas as pd
from torch.utils.data import Dataset
from typing import Optional
import torch
from .image_text_config import ImageTextClassifierConfig
from ..base_datamodule import BaseDataModule


class ImageTextClassifierDataset(Dataset):
    """
    Image Text Classifier Dataset.

    Args:
        data (pd.DataFrame): Dataframe with columns ["img_path", "text", "label"].
        problem_type (str): Problem type, either "multi_label_classification" or "single_label_classification".
        num_class (int): Number of classes.
    """

    def __init__(self, data: pd.DataFrame, problem_type: str, num_class: int):
        super().__init__()
        self.img_path_list = data["img_path"].tolist()
        self.text_list = data["text"].tolist()
        self.label_list = data["label"].tolist()
        self.problem_type = problem_type
        self.num_class = num_class

    def __getitem__(self, idx):
        input_img_path = self.img_path_list[idx]
        input_text = self.text_list[idx]
        label = self.label_list[idx]

        if self.problem_type == "multi_label_classification":
            target = torch.zeros(self.num_class)
            # Case for single label, but want to use multi-label loss (e.g. BCE)
            if type(label) != list:
                label = [label]
            for lab in label:
                # negative sample
                if lab == -1:
                    continue
                target[lab] = 1

            return input_img_path, input_text, target
        else:
            return input_img_path, input_text, label

    def __len__(self):
        return len(self.label_list)


class ImageTextClassifierDataModule(BaseDataModule):
    """
    Image Text Classifier Data Module.

    Args:
        train_dataset (pd.DataFrame): Training dataset.
        val_dataset (pd.DataFrame): Validation dataset.
        test_dataset (pd.DataFrame): Test dataset.
        train_config (ImageTextClassifierConfig): ImageTextClassifierConfig.
    """

    def __init__(
        self,
        train_dataset: pd.DataFrame,
        val_dataset: pd.DataFrame,
        test_dataset: pd.DataFrame,
        train_config: ImageTextClassifierConfig,
    ):
        super().__init__(train_dataset, val_dataset, test_dataset, train_config)

    def setup(self, stage: Optional[str] = None):
        self.trainset = ImageTextClassifierDataset(
            self.train_dataset, self.problem_type, self.num_class
        )
        self.valset = ImageTextClassifierDataset(
            self.val_dataset, self.problem_type, self.num_class
        )
        self.testset = ImageTextClassifierDataset(
            self.test_dataset, self.problem_type, self.num_class
        )


class FeedbackImageTextClassifierDataset(Dataset):
    """
    Feedback ImageText Classifier Dataset.

    Args:
        data (pd.DataFrame): Dataframe containing image path and label.
        problem_type (str): Problem type, either "single_label_classification" or "multi_label_classification".
        num_class (int): Number of classes.
    """

    def __init__(self, data: pd.DataFrame, problem_type: str, num_class: int):
        super().__init__()
        self.img_path_list = data["img_path"].tolist()
        self.text_list = data["text"].tolist()
        self.label_list = data["label"].tolist()
        self.feedback_flag_list = data["feedback_flag"].tolist()
        self.problem_type = problem_type
        self.num_class = num_class

    def __getitem__(self, idx):
        input_img_path = self.img_path_list[idx]
        input_text = self.text_list[idx]
        label = self.label_list[idx]
        feedback_flag = self.feedback_flag_list[idx]

        if self.problem_type == "multi_label_classification":
            target = torch.zeros(self.num_class)
            # Case for single label, but want to use multi-label loss (e.g. BCE)
            if type(label) != list:
                label = [label]
            for lab in label:
                # negative sample
                if lab == -1:
                    continue
                target[lab] = 1

            return input_img_path, input_text, target, feedback_flag
        # single label classification
        else:
            return input_img_path, input_text, label, feedback_flag

    def __len__(self):
        return len(self.label_list)



class FeedbackImageTextClassifierDataModule(BaseDataModule):
    """
    Feedback ImageText Classifier Data Module.

    Args:
        train_dataset (pd.DataFrame): Training dataset (containing original train data and feedback data).
        val_dataset (pd.DataFrame): Validation dataset.
        test_dataset (pd.DataFrame): Test dataset.
        train_config (ImageClassifierConfig): ImageClassifierConfig.
    """

    def __init__(
        self,
        train_dataset: pd.DataFrame,
        val_dataset: pd.DataFrame,
        test_dataset: pd.DataFrame,
        train_config: ImageTextClassifierConfig,
    ):
        super().__init__(train_dataset, val_dataset, test_dataset, train_config)

    def setup(self, stage: Optional[str] = None):
        self.trainset = FeedbackImageTextClassifierDataset(
            self.train_dataset, self.problem_type, self.num_class
        )
        # Use same dataset for validation with and without feedback
        self.valset = ImageTextClassifierDataset(
            self.val_dataset, self.problem_type, self.num_class
        )
        # Use same dataset for test with and without feedback
        self.testset = ImageTextClassifierDataset(
            self.test_dataset, self.problem_type, self.num_class
        )


import logging
from ..base_config import BaseClassifierConfig

logger = logging.getLogger(__name__)


class ImageClassifierConfig(BaseClassifierConfig):
    """
    Image Classifier Config.
    Containing all the hyperparameters for training and inference.

    Args:
        backbone_name (str): Name of the backbone model. Defaults to "eva_large_patch14_196.in22k_ft_in22k_in1k".
        classifier_learning_rate (float): Learning rate for the classifier. Defaults to 1e-3.
        backbone_learning_rate (float): Learning rate for the backbone. Defaults to 4e-5.
        freeze (bool): Freeze the backbone. Defaults to False.
        lora (bool): Use LoRA. Defaults to False.
        lora_r (int): LoRA r. Defaults to 16.
        lora_alpha (int): LoRA alpha. Defaults to 16.
        lora_target_modules (list): LoRA target modules. Defaults to ["qkv"].
        lora_dropout (float): LoRA dropout. Defaults to 0.1.
        auto_augment (str): Image augmentation method. Defaults to "rand-n3-mstd1".
        crop_ratio (float): Minimum crop ratio for when augmentate image, default: 0.9.
    """

    def __init__(
        self,
        backbone_name: str = "eva_large_patch14_196.in22k_ft_in22k_in1k",
        classifier_learning_rate: float = 1e-3,
        backbone_learning_rate: float = 4e-5,
        freeze: bool = False,
        lora: bool = False,
        lora_r: int = 16,
        lora_alpha: int = 16,
        lora_target_modules: list = None,
        lora_dropout: float = 0.1,
        auto_augment: str = "rand-n3-mstd1",
        crop_ratio: float = 0.9,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.backbone_name = backbone_name
        self.classifier_learning_rate = classifier_learning_rate
        self.backbone_learning_rate = backbone_learning_rate
        self.freeze = freeze
        self.lora = lora
        if self.freeze and self.lora:
            raise ValueError(
                "freezing backbone and using lora cannot be True at the same time"
            )

        self.lora_r = lora_r
        self.lora_alpha = lora_alpha
        if lora_target_modules is None:
            self.lora_target_modules = ["qkv"]
        else:
            self.lora_target_modules = lora_target_modules
        self.lora_dropout = lora_dropout

        # Image augmentation method
        self.auto_augment = auto_augment
        self.crop_ratio = crop_ratio
        # Boolean to check if the model is loaded from transformers or timm, will be set in trainer
        self.timm = None



import torch
import torch.nn.functional as F
from typing import Union


def focal_loss(
    inputs: torch.Tensor,
    targets: torch.Tensor,
    gamma: float = 2.0,
    alpha: Union[float, int, list] = None,
    reduction: str = "mean",
    **kwargs
):
    """
    Focal Loss for unblanced classification problem.
    Args:
        inputs (torch.Tensor): model outputs
        targets (torch.Tensor): ground truth labels
        gamma (float): focal loss gamma
        alpha (Union[float, int, list]): alpha for focal loss
        reduction (str): reduction method
    Returns:
        torch.Tensor: focal loss
    """
    # Calculate the cross-entropy loss
    CE_loss = F.cross_entropy(inputs, targets, reduction="none")

    # Convert CE_loss to probabilities
    pt = torch.exp(-CE_loss)

    # Calculate the Focal Loss
    F_loss = (1 - pt) ** gamma * CE_loss

    if alpha is not None:
        if isinstance(alpha, (float, int)):
            alpha = torch.Tensor([1 - alpha, alpha]).to(inputs.device)
        elif isinstance(alpha, list):
            alpha = torch.Tensor(alpha).to(inputs.device)
        else:
            raise ValueError("alpha should be float, int or list")
        at = alpha.gather(0, targets)
        F_loss = at * F_loss

    if reduction == "mean":
        return F_loss.mean()
    elif reduction == "sum":
        return F_loss.sum()
    else:
        return F_loss


def binary_focal_loss(
    inputs: torch.Tensor,
    targets: torch.Tensor,
    gamma: float = 2.0,
    alpha: Union[float, int] = None,
    reduction: str = "mean",
    **kwargs
):
    """
    Focal Loss for binary classification problem.
    Args:
        inputs (torch.Tensor): model outputs
        targets (torch.Tensor): ground truth labels
        gamma (float): focal loss gamma
        alpha (Union[float, int]): alpha for focal loss
        reduction (str): reduction method
    Returns:
        torch.Tensor: binary focal loss
    """
    # Calculate the cross-entropy loss
    CE_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")

    # Convert CE_loss to probabilities
    pt = torch.exp(-CE_loss)

    # Calculate the Focal Loss
    F_loss = (1 - pt) ** gamma * CE_loss

    if alpha is not None:
        if isinstance(alpha, (float, int)):
            alpha = torch.Tensor([1 - alpha, alpha]).to(inputs.device)
        else:
            raise ValueError("alpha should be float or int")
        at = alpha.gather(0, targets)
        F_loss = at * F_loss

    if reduction == "mean":
        return F_loss.mean()
    elif reduction == "sum":
        return F_loss.sum()
    else:
        return F_loss


def focal_loss_with_label_smoothing(
    inputs: torch.Tensor,
    targets: torch.Tensor,
    gamma: float = 2.0,
    alpha: Union[float, int, list] = None,
    reduction: str = "mean",
    smoothing: float = 0.1,
    num_classes: int = None,
    **kwargs
):
    """
    Focal Loss with label smoothing for unbalanced classification problem.
    Args:
        inputs (torch.Tensor): model outputs
        targets (torch.Tensor): ground truth labels
        gamma (float): focal loss gamma
        alpha (Union[float, int, list]): alpha for focal loss
        reduction (str): reduction method
        smoothing (float): label smoothing factor
        num_classes (int): number of classes
    Returns:
        torch.Tensor: focal loss with label smoothing
    """

    # Ensure num_classes is provided
    if num_classes is None:
        num_classes = inputs.shape[-1]

    # Calculate cross-entropy with label smoothing
    targets = targets.view(-1, 1)
    log_probs = F.log_softmax(inputs, dim=1)
    targets_one_hot = torch.zeros_like(log_probs).scatter(1, targets, 1)
    targets_smooth = (1.0 - smoothing) * targets_one_hot + smoothing / num_classes
    CE_loss = -torch.sum(targets_smooth * log_probs, dim=1)

    # Focal loss calculations
    pt = log_probs.gather(1, targets).view(-1).exp()

    F_loss = (1 - pt) ** gamma * CE_loss
    if alpha is not None:
        if isinstance(alpha, (float, int)):
            alpha = torch.Tensor([alpha, 1 - alpha]).to(inputs.device)
        elif isinstance(alpha, list):
            alpha = torch.Tensor(alpha).to(inputs.device)
        at = alpha.gather(0, targets.view(-1))
        F_loss = at * F_loss

    if reduction == "mean":
        return F_loss.mean()
    elif reduction == "sum":
        return F_loss.sum()
    else:
        return F_loss


def bce_with_logits(inputs: torch.Tensor, targets: torch.Tensor, **kwargs):
    """
    Binary cross entropy with logits
    Args:
        inputs (torch.Tensor): model outputs
        targets (torch.Tensor): ground truth labels
    Returns:
        torch.Tensor: binary cross entropy with logits
    """
    return F.binary_cross_entropy_with_logits(
        inputs, targets
    )  # , pos_weight=pos_weight.to(inputs.device))


def cross_entropy_from_probabilities(true_distributions, pred_distributions):
    """
    Calculates the cross entropy between two probability distributions that have already been applied with softmax.

    Parameters:
    - true_distributions: The actual distribution (target distribution). A PyTorch tensor.
    - pred_distributions: The predicted distribution. A PyTorch tensor.

    Returns:
    - The cross entropy value between the two distributions. (scalar)
    """
    # 두 분포 간의 cross entropy 계산
    cross_entropy = -torch.sum(
        true_distributions * torch.log(pred_distributions + 1e-9)
    )  # / true_distributions.size(0)
    return cross_entropy


def binary_cross_entropy_from_probabilities(true_distributions, pred_distributions):
    """
    두 확률 분포 사이의 이진 교차 엔트로피를 계산합니다. 이 확률 분포는 이미 softmax가 적용된 상태입니다.

    매개변수:
    - true_distributions: 실제 분포 (목표 분포). PyTorch 텐서.
    - pred_distributions: 예측 분포. PyTorch 텐서.

    반환값:
    - 두 분포 간의 이진 교차 엔트로피 값. (스칼라)
    """
    # 정밀도 문제를 피하기 위해 float32로 변환
    true_distributions = true_distributions.to(torch.float32)
    pred_distributions = pred_distributions.to(torch.float32)

    # 로그의 입력값이 0이 되는 것을 방지하기 위해 작은 epsilon 추가
    epsilon = 1e-4
    pred_distributions = torch.clamp(pred_distributions, epsilon, 1.0 - epsilon)

    # 이진 교차 엔트로피 계산
    binary_cross_entropy = -torch.sum(
        true_distributions * torch.log(pred_distributions)
        + (1 - true_distributions) * torch.log(1 - pred_distributions)
    )

    return binary_cross_entropy


LOSSES = {
    "cross_entropy": F.cross_entropy,
    "focal_loss": focal_loss,
    "focal_loss_with_label_smoothing": focal_loss_with_label_smoothing,
    "binary_cross_entropy": bce_with_logits,
    "binary_focal_loss": binary_focal_loss,
}



import inspect
import logging
import traceback
import timm
from typing import List

import pytorch_lightning as pl
import torch
import torch.nn as nn
from peft import LoraConfig, get_peft_model
from transformers import get_linear_schedule_with_warmup, AutoConfig
from torchvision import transforms

from .head_classifier import *
from .trainer_util import (
    TextClassifierInputPreprocess,
    TextClassifierSelectToken,
    ImageClassifierInputPreprocess,
    TransformerWrapper,
    SquarePad,
    _str_to_pil_interpolation,
    _pil_interpolation_to_str,
    ModelEMA,
    find_indices_based_on_max_and_label,
)
from ..losses import (
    LOSSES,
    cross_entropy_from_probabilities,
    binary_cross_entropy_from_probabilities,
)
from torch.nn import functional as F

logger = logging.getLogger(__name__)


class FeedbackBaseClassifierModule(pl.LightningModule):
    """
    Base class for all classifier modules

    Args:
        train_config (TrainConfig): train config
    """

    def __init__(self, train_config):
        super().__init__()

        self.train_config = train_config
        self.criterion = LOSSES[train_config.criterion]

        self.automatic_optimization = False
        self.train_step_outputs = []
        self.validation_step_outputs = []
        self.test_step_outputs = []
        self.test_result = []
        self.ema_backbone = None

    def on_train_epoch_start(self) -> None:
        from .image_text_classifier.image_text_config import ImageTextClassifierConfig

        if type(self.train_config) == ImageTextClassifierConfig:
            self.text_backbone.train()
            self.image_backbone.train()
            self.backbone_list = nn.ModuleList(
                [self.image_backbone, self.text_backbone]
            )
        else:
            self.backbone.train()
            self.backbone_list = nn.ModuleList([self.backbone])
        self.ema_backbone = ModelEMA(
            self.backbone_list, self.train_config.feedback_ema_decay
        )
        for ema_backbone in self.ema_backbone.ema:
            ema_backbone.eval()
        self.ema_backbone.updates = 0

        self.classifier_head.train()
        self.ema_classifier_head = ModelEMA(
            (self.classifier_head,), self.train_config.feedback_ema_decay
        )
        for ema_classifier in self.ema_classifier_head.ema:
            ema_classifier.eval()
        self.ema_classifier_head.updates = 0

    def get_model(self):
        raise NotImplementedError

    def forward(self):
        raise NotImplementedError

    def configure_optimizers(self):
        raise NotImplementedError

    ## Bugfixed in peft 0.7.0 (so we cast lora weights to float, for mixed precision training)
    def cast_lora_to_float(self, model: nn.Module):
        for name, mod in model.named_modules():
            if ("lora_" in name) and hasattr(mod, "weight"):
                mod.weight.data = mod.weight.data.float()
            if ("lora_" in name) and hasattr(mod, "bias") and (mod.bias is not None):
                mod.bias.data = mod.bias.data.float()
            if ("modules_to_save" in name) and isinstance(mod, nn.Linear):
                mod.weight.data = mod.weight.data.float()
                if mod.bias is not None:
                    mod.bias.data = mod.bias.data.float()

    def load_model(
        self,
        model: nn.Module,
        freeze: bool,
        lora: bool,
        lora_r: float,
        lora_alpha: float,
        lora_target_modules: List[str],
        lora_dropout: float,
    ):
        """
        Load model and set up LORA if necessary

        Args:
            model (nn.Module): model
            freeze (bool): whether to freeze the backbone
            lora (bool): whether to use LORA
            lora_r (float): LORA r parameter
            lora_alpha (float): LORA alpha parameter
            lora_target_modules (list with string): LORA target modules
            lora_dropout (float): LORA dropout parameter
        """
        # Freeze
        if freeze:
            logger.info("Fine-tuning while backbone freezed.")
            for param in model.parameters():
                param.requires_grad = False
            model.eval()
        # LORA finetuning
        elif lora:
            logger.info("Fine-tuning with LORA")
            lora_config = LoraConfig(
                r=lora_r,
                lora_alpha=lora_alpha,
                target_modules=lora_target_modules,
                lora_dropout=lora_dropout,
                modules_to_save=["classifier"],
            )
            try:
                model = get_peft_model(model, lora_config)
                self.cast_lora_to_float(model)
            except ValueError:
                raise ValueError(
                    f"Error message: {traceback.format_exc()}\n\nLORA failed, this may be due to the mismatch between lora_target_modules={lora_target_modules} and text model structure.\nThe structure of the text model is:\n{model}"
                )
        # Full fine-tuning
        else:
            logger.warning(
                "You are trying to train model without freezing the backbone or using LORA. You can encounter out of memory error easily."
            )

        return model

    def get_text_model(
        self,
        backbone_name: str,
        freeze: bool,
        lora: bool,
        lora_r: float,
        lora_alpha: float,
        lora_target_modules: List[str],
        lora_dropout: float,
    ):
        """
        Load text model and tokenizer

        Args:
            backbone_name (str): backbone name
            freeze (bool): whether to freeze the backbone
            lora (bool): whether to use LORA
            lora_r (float): LORA r parameter
            lora_alpha (float): LORA alpha parameter
            lora_target_modules (list with string): LORA target modules
            lora_dropout (float): LORA dropout parameter

        """
        from transformers import AutoConfig, AutoModel, AutoTokenizer

        logger.info("Loading pretrained model...")
        model = AutoModel.from_pretrained(
            backbone_name, low_cpu_mem_usage=True, torch_dtype="auto"
        )
        tokenizer = AutoTokenizer.from_pretrained(backbone_name)
        config = AutoConfig.from_pretrained(backbone_name)
        # # Save initial config for future use.
        # self.initial_config = config
        # self.embedding_size = config.hidden_size

        # Check unnecssary input keys
        forward_parameters = inspect.signature(model.forward).parameters
        delete_input_key_list = []
        for input_key in tokenizer.model_input_names:
            if input_key not in forward_parameters.keys():
                delete_input_key_list.append(input_key)
                logger.info(
                    f"{input_key} not needed for model forward, will be deleted automatically."
                )
        model = self.load_model(
            model, freeze, lora, lora_r, lora_alpha, lora_target_modules, lora_dropout
        )

        return model, tokenizer, config, delete_input_key_list

    def get_image_model(
        self,
        backbone_name: str,
        freeze: bool,
        lora: bool,
        lora_r: float,
        lora_alpha: float,
        lora_target_modules: List[str],
        lora_dropout: float,
    ):
        """
        Load image model

        Args:
            backbone_name (str): backbone name
            freeze (bool): whether to freeze the backbone
            lora (bool): whether to use LORA
            lora_r (float): LORA r parameter
            lora_alpha (float): LORA alpha parameter
            lora_target_modules (list with string): LORA target modules
            lora_dropout (float): LORA dropout parameter
        """
        if backbone_name in timm.list_models(pretrained=True):
            model = timm.create_model(
                backbone_name, pretrained=True, num_classes=0
            )  # remove linear classifier

            data_config = timm.data.resolve_model_data_config(model)
            train_transforms = transforms.Compose(
                [
                    SquarePad(),
                    transforms.Resize(
                        data_config["input_size"][-2:],
                        interpolation=_str_to_pil_interpolation[
                            data_config["interpolation"]
                        ],
                    ),
                    transforms.RandomHorizontalFlip(p=0.5),
                    transforms.ColorJitter(
                        brightness=(0.6, 1.4),
                        contrast=(0.6, 1.4),
                        saturation=(0.6, 1.4),
                        hue=0,
                    ),
                    transforms.ToTensor(),
                    transforms.Normalize(data_config["mean"], data_config["std"]),
                ]
            )

            val_transforms = transforms.Compose(
                [
                    SquarePad(),
                    transforms.Resize(
                        data_config["input_size"][-2:],
                        interpolation=_str_to_pil_interpolation[
                            data_config["interpolation"]
                        ],
                    ),
                    transforms.ToTensor(),
                    transforms.Normalize(data_config["mean"], data_config["std"]),
                ]
            )
            timm_flag = True
            config = None
            # self.train_config.timm = True
        else:
            timm_flag = False

            from transformers import (
                AutoImageProcessor,
                AutoModel,
                CLIPModel,
                AutoConfig,
            )

            model = AutoModel.from_pretrained(backbone_name)
            if isinstance(model, CLIPModel):
                model = model.vision_model
                # Cannot converted with huggingface
                timm_flag = True
            model = TransformerWrapper(model)
            image_processor = AutoImageProcessor.from_pretrained(backbone_name)
            config = AutoConfig.from_pretrained(backbone_name)

            if hasattr(image_processor, "crop_size"):
                data_config = {
                    "input_size": (
                        3,
                        image_processor.crop_size["height"],
                        image_processor.crop_size["width"],
                    ),
                    "mean": image_processor.image_mean,
                    "std": image_processor.image_std,
                    "interpolation": _pil_interpolation_to_str[
                        image_processor.resample
                    ],
                    "scale": (self.train_config.crop_ratio, 1.0),
                }
            elif hasattr(image_processor, "size"):
                data_config = {
                    "input_size": (
                        3,
                        image_processor.size["height"],
                        image_processor.size["width"],
                    ),
                    "mean": image_processor.image_mean,
                    "std": image_processor.image_std,
                    "interpolation": _pil_interpolation_to_str[
                        image_processor.resample
                    ],
                    "scale": (self.train_config.crop_ratio, 1.0),
                }
            else:
                raise ValueError(
                    f"'size' and 'crop_size' does not exist in current Image processor. Current image processor: {image_processor}"
                )

            train_transforms = transforms.Compose(
                [
                    SquarePad(),
                    transforms.Resize(
                        data_config["input_size"][-2:],
                        interpolation=_str_to_pil_interpolation[
                            data_config["interpolation"]
                        ],
                    ),
                    transforms.RandomHorizontalFlip(p=0.5),
                    transforms.ColorJitter(
                        brightness=(0.6, 1.4),
                        contrast=(0.6, 1.4),
                        saturation=(0.6, 1.4),
                        hue=0,
                    ),
                    transforms.ToTensor(),
                    transforms.Normalize(data_config["mean"], data_config["std"]),
                ]
            )

            val_transforms = transforms.Compose(
                [
                    SquarePad(),
                    transforms.Resize(
                        data_config["input_size"][-2:],
                        interpolation=_str_to_pil_interpolation[
                            data_config["interpolation"]
                        ],
                    ),
                    transforms.ToTensor(),
                    transforms.Normalize(data_config["mean"], data_config["std"]),
                ]
            )

        model = self.load_model(
            model, freeze, lora, lora_r, lora_alpha, lora_target_modules, lora_dropout
        )

        return model, train_transforms, val_transforms, data_config, timm_flag, config

    def text_forward(
        self,
        input_texts: List[str],
        backbone: nn.Module,
        classifier_head: nn.Module,
        initial_config: AutoConfig,
        freeze: bool = False,
        eval: bool = False,
        classifier: bool = True,
    ):
        """
        Forward function for text model

        Args:
            input_texts (List[str]): list of input texts
            backbone (nn.Module): backbone model
            eval (bool): whether to use eval mode
            freeze (bool): whether to freeze the backbone
            classifier (bool): whether to use classifier head
            initial_config (AutoConfig): initial config
        """
        inputs = TextClassifierInputPreprocess(
            input_texts,
            text_prompt=self.train_config.text_prompt,
            device=self.device,
            tokenizer=self.tokenizer,
        )

        # Delete unnecessary input keys
        for delete_input_key in self.delete_input_key_list:
            del inputs[delete_input_key]

        # For classificaiton, model use the last token as other causal models (e.g. GPT-1) do. We utilize attention mask to know the position of the last token.
        if eval:
            with torch.no_grad():
                backbone_outputs = backbone(**inputs)
                backbone_outputs = TextClassifierSelectToken(
                    backbone_outputs, inputs=inputs, configs=initial_config
                )
                if not classifier:
                    return backbone_outputs
                logits = classifier_head(backbone_outputs)
            return logits
        else:
            if freeze:
                # Do not compute gradient for backbone weights
                with torch.no_grad():
                    backbone_outputs = backbone(**inputs)
                    backbone_outputs = TextClassifierSelectToken(
                        backbone_outputs, inputs=inputs, configs=initial_config
                    )
            # LORA and full fine-tuning case.
            else:
                backbone_outputs = backbone(**inputs)
                backbone_outputs = TextClassifierSelectToken(
                    backbone_outputs, inputs=inputs, configs=initial_config
                )

            if not classifier:
                return backbone_outputs
            logits = classifier_head(backbone_outputs)

            return logits

    def image_forward(
        self,
        input_img_paths: List[str],
        backbone: nn.Module,
        classifier_head: nn.Module,
        freeze: bool = False,
        eval: bool = False,
        classifier: bool = True,
    ):
        """
        Forward function for image model

        Args:
            input_img_paths (List[str]): input image paths
            backbone (nn.Module): backbone model
            eval (bool): whether to use eval mode
            freeze (bool): whether to freeze the backbone
            classifier (bool): whether to use classifier head
        """
        if eval:
            inputs = ImageClassifierInputPreprocess(
                input_img_paths,
                device=self.device,
                transforms=self.val_transforms,
            )
            with torch.no_grad():
                backbone_outputs = backbone.forward(inputs)
                if not classifier:
                    return backbone_outputs
                logits = classifier_head(backbone_outputs)
            return logits
        else:
            inputs = ImageClassifierInputPreprocess(
                input_img_paths,
                device=self.device,
                transforms=self.train_transforms,
            )
            if freeze:
                # Do not compute gradient for backbone weights
                with torch.no_grad():
                    backbone_outputs = backbone.forward(inputs)
            # LORA and full fine-tuning case.
            else:
                backbone_outputs = backbone.forward(inputs)

            if not classifier:
                return backbone_outputs

            logits = classifier_head(backbone_outputs)

            return logits

    def training_step(self, batch, batch_idx: int):
        """
        Training step for LightningModule
        """

        self.ema_backbone.update(self.backbone_list)

        self.ema_classifier_head.update((self.classifier_head,))
        feedback_flags = batch[-1]
        labels = batch[-2]
        inputs = batch[:-2]
        if self.train_config.problem_type == "single_label_classification":
            labels = labels.type(torch.int64)
        logits = self.forward(*inputs)
        feedback_idx = [
            index for index, value in enumerate(feedback_flags) if value == 1
        ]
        non_feedback_idx = [
            index for index, value in enumerate(feedback_flags) if value == 0
        ]
        non_feedback_inputs = [
            [sub_inputs[i] for i in non_feedback_idx] for sub_inputs in inputs
        ]
        feedback_inputs = [
            [sub_inputs[i] for i in feedback_idx] for sub_inputs in inputs
        ]
        # For feedback data, we use self-play finetuning loss. https://arxiv.org/pdf/2401.01335.pdf (Zixiang Chen)
        if feedback_idx:
            if self.train_config.problem_type == "single_label_classification":
                feedback_logits = logits[feedback_idx]
                feedback_labels = labels[feedback_idx]
                feedback_prob = F.softmax(feedback_logits / 1.0, dim=1)
                true_prob = feedback_prob[
                    torch.arange(len(feedback_labels)), feedback_labels
                ]
                pred_prob = torch.max(feedback_prob, dim=1).values

                pseudo_feedback_logits = self.forward(
                    *feedback_inputs,
                    *self.ema_backbone.ema,
                    *self.ema_classifier_head.ema,
                    eval=True,
                )
                pseudo_feedback_prob = F.softmax(pseudo_feedback_logits / 1.0, dim=1)
                pseudo_true_prob = pseudo_feedback_prob[
                    torch.arange(len(feedback_labels)), feedback_labels
                ]
                pseudo_pred_prob = torch.max(pseudo_feedback_prob, dim=1).values

                feedback_loss = torch.log(
                    1
                    + torch.exp(
                        -(
                            torch.log(true_prob / (pseudo_true_prob + 1e-9) + 1e-9)
                            - torch.log(pred_prob / (pseudo_pred_prob + 1e-9) + 1e-9)
                        )
                    )
                )
                feedback_loss = torch.sum(feedback_loss)
            else:
                feedback_logits = logits[feedback_idx]
                feedback_labels = labels[feedback_idx]
                feedback_prob = F.sigmoid(feedback_logits)
                true_prob = (1 - feedback_labels) * (
                    1 - feedback_prob
                ) + feedback_labels * feedback_prob
                pred_prob = torch.abs(feedback_prob - 0.5) + 0.5

                pseudo_feedback_logits = self.forward(
                    *feedback_inputs,
                    *self.ema_backbone.ema,
                    *self.ema_classifier_head.ema,
                    eval=True,
                )
                pseudo_feedback_prob = F.sigmoid(pseudo_feedback_logits)
                pseudo_true_prob = (1 - feedback_labels) * (
                    1 - pseudo_feedback_prob
                ) + feedback_labels * pseudo_feedback_prob
                pseudo_pred_prob = torch.abs(pseudo_feedback_prob - 0.5) + 0.5

                feedback_loss = torch.log(
                    1
                    + torch.exp(
                        -(
                            torch.log(true_prob / (pseudo_true_prob + 1e-9) + 1e-9)
                            - torch.log(pred_prob / (pseudo_pred_prob + 1e-9) + 1e-9)
                        )
                    )
                )
                feedback_loss = torch.sum(feedback_loss)

        else:
            feedback_loss = 0

        # For non-feedback data(=sample from trai data), we use cross entropy loss(or focal loss, depend on user-defined criterion) & Knowledge distillation loss. only use cross entropy loss
        # for high confidence samples (e.g. softmax > 0.7) and same predictions and labels.
        if non_feedback_idx:
            if self.train_config.problem_type == "single_label_classification":
                non_feedback_logits = logits[non_feedback_idx]
                softmax_non_feedback_logits = F.softmax(
                    non_feedback_logits / 1.0, dim=1
                )

                higher_idx, lower_idx = find_indices_based_on_max_and_label(
                    softmax_non_feedback_logits, labels[non_feedback_idx], threshold=0.7
                )

                higher_cross_entropy_loss = self.criterion(
                    non_feedback_logits[higher_idx],
                    labels[non_feedback_idx][higher_idx],
                    label_smoothing=self.train_config.label_smoothing_factor,
                    reduction="sum",
                )

                lower_non_feedback_inputs = [
                    [sub_inputs[i] for i in lower_idx]
                    for sub_inputs in non_feedback_inputs
                ]
                if len(lower_idx) > 0:
                    non_feedback_logits = self.forward(
                        *lower_non_feedback_inputs,
                        *self.ema_backbone.ema,
                        *self.ema_classifier_head.ema,
                        eval=True,
                    )

                    softmax_pseudo_label_logit = F.softmax(
                        non_feedback_logits / 1.0, dim=1
                    )

                    # Knowledge distillation losss
                    lower_kd_loss = cross_entropy_from_probabilities(
                        softmax_pseudo_label_logit,
                        softmax_non_feedback_logits[lower_idx],
                    )

                else:
                    lower_kd_loss = 0

            else:  # multi-label classification
                non_feedback_logits = logits[non_feedback_idx]
                non_feedback_labels = labels[non_feedback_idx]
                non_feedback_prob = F.sigmoid(non_feedback_logits)

                # flattening for simple calculation
                non_feedback_prob = non_feedback_prob.view(-1)
                non_feedback_labels = non_feedback_labels.view(-1)

                ##### 요 부분 조절 필요!!! @김태윤 ######
                higher_idx = torch.where(
                    torch.abs(non_feedback_prob - non_feedback_labels) <= 0.1
                )[0]
                lower_idx = torch.where(
                    torch.abs(non_feedback_prob - non_feedback_labels) > 0.1
                )[0]

                higher_cross_entropy_loss = self.criterion(
                    non_feedback_logits.view(-1)[higher_idx],
                    non_feedback_labels[higher_idx],
                    label_smoothing=self.train_config.label_smoothing_factor,
                    reduction="sum",
                )

                if len(lower_idx) > 0:
                    non_feedback_pseudo_logits = self.forward(
                        *non_feedback_inputs,
                        *self.ema_backbone.ema,
                        *self.ema_classifier_head.ema,
                        eval=True,
                    )
                    pseudo_non_feedback_prob = F.sigmoid(non_feedback_pseudo_logits)
                    pseudo_non_feedback_prob = pseudo_non_feedback_prob.view(-1)

                    lower_kd_loss = binary_cross_entropy_from_probabilities(
                        pseudo_non_feedback_prob[lower_idx],
                        non_feedback_prob[lower_idx],
                    )

                else:
                    lower_kd_loss = 0

            non_feedback_loss = lower_kd_loss + higher_cross_entropy_loss

        else:
            non_feedback_loss = 0

        loss = (feedback_loss + non_feedback_loss) / len(labels)
        ## loss가 nan인지 확인
        if torch.isnan(loss):
            breakpoint()

        self.log("train/loss", loss.item(), prog_bar=True)

        # Manual backward
        opt = self.optimizers()
        sch = self.lr_schedulers()
        sch.step()
        self.manual_backward(loss)

        if self.train_config.gradient_accumulation_steps > 1:
            # Apply gradient accumulation
            if (batch_idx + 1) % self.train_config.gradient_accumulation_step == 0:
                opt.step()
                opt.zero_grad()
        else:
            opt.step()
            opt.zero_grad()

        if self.train_config.problem_type == "single_label_classification":
            _, prediction = torch.max(logits, 1)
            correct_pred = (prediction == labels).sum().item()
            total_pred = len(prediction)
            self.train_step_outputs.append(
                {"correct_pred": correct_pred, "total_pred": total_pred}
            )

        else:
            preds = torch.where(
                logits > 0, torch.ones_like(logits), torch.zeros_like(logits)
            )
            TP = (preds * labels).sum(dim=0)
            FP = (preds * (1 - labels)).sum(dim=0)
            FN = ((1 - preds) * labels).sum(dim=0)
            TN = ((1 - preds) * (1 - labels)).sum(dim=0)

            self.train_step_outputs.append({"TP": TP, "FP": FP, "FN": FN, "TN": TN})

    @torch.no_grad()
    def validation_step(self, batch, batch_idx: int):
        """
        Validation step for LightningModule
        """
        labels = batch[-1]
        inputs = batch[:-1]
        # input_img_paths, labels = batch
        if self.train_config.problem_type == "single_label_classification":
            labels = labels.type(torch.int64)

        if self.ema_backbone is None:
            logits = self.forward(*inputs, eval=True)
        else:
            logits = self.forward(
                *inputs,
                *self.ema_backbone.ema,
                *self.ema_classifier_head.ema,
                eval=True,
            )

        val_loss = self.criterion(
            logits, labels, label_smoothing=self.train_config.label_smoothing_factor
        )
        self.log("val/loss", val_loss.item(), prog_bar=True)

        if self.train_config.problem_type == "single_label_classification":
            # Calculate for topk accuracy
            topk_list = self.train_config.print_topk_list
            correct_pred_list = []
            for topk in topk_list:
                _, predictions = logits.topk(topk, 1, True, True)
                correct_pred = predictions.eq(labels.view(-1, 1)).view(-1).sum().item()
                correct_pred_list.append(correct_pred)

            total_pred = len(predictions)

            self.validation_step_outputs.append(
                {"correct_pred_list": correct_pred_list, "total_pred": total_pred}
            )
        # Multi-label classification
        else:
            preds = torch.where(
                logits > 0, torch.ones_like(logits), torch.zeros_like(logits)
            )
            TP = (preds * labels).sum(dim=0)
            FP = (preds * (1 - labels)).sum(dim=0)
            FN = ((1 - preds) * labels).sum(dim=0)
            TN = ((1 - preds) * (1 - labels)).sum(dim=0)

            self.validation_step_outputs.append(
                {"TP": TP, "FP": FP, "FN": FN, "TN": TN}
            )

    def test_step(self, batch, batch_idx: int):
        """
        Test step for LightningModule
        """
        labels = batch[-1]
        inputs = batch[:-1]
        # input_img_paths, labels = batch
        if self.train_config.problem_type == "single_label_classification":
            labels = labels.type(torch.int64)

        if self.ema_backbone is None:
            logits = self.forward(*inputs, eval=True)
        else:
            logits = self.forward(
                *inputs,
                *self.ema_backbone.ema,
                *self.ema_classifier_head.ema,
                eval=True,
            )

        # val_loss = F.cross_entropy(
        #     logits, labels, label_smoothing=self.train_config.label_smoothing_factor
        # )
        test_loss = self.criterion(
            logits, labels, label_smoothing=self.train_config.label_smoothing_factor
        )
        self.log("test/loss", test_loss.item(), prog_bar=True, sync_dist=True)
        if self.train_config.problem_type == "single_label_classification":
            # Calculate for topk accuracy
            topk_list = self.train_config.print_topk_list
            correct_pred_list = []
            for topk in topk_list:
                _, predictions = logits.topk(topk, 1, True, True)
                correct_pred = predictions.eq(labels.view(-1, 1)).view(-1).sum().item()
                correct_pred_list.append(correct_pred)

            total_pred = len(predictions)

            self.test_step_outputs.append(
                {"correct_pred_list": correct_pred_list, "total_pred": total_pred}
            )
        # Multi-label classification
        else:
            preds = torch.where(
                logits > 0, torch.ones_like(logits), torch.zeros_like(logits)
            )
            TP = (preds * labels).sum(dim=0)
            FP = (preds * (1 - labels)).sum(dim=0)
            FN = ((1 - preds) * labels).sum(dim=0)
            TN = ((1 - preds) * (1 - labels)).sum(dim=0)

            self.test_step_outputs.append({"TP": TP, "FP": FP, "FN": FN, "TN": TN})

    def on_train_epoch_end(self):
        """
        Called at the end of training epoch, here we log the train accuracy or other metrics e.g. f1 score
        """
        # self.ema_backbone.update_attr(self.backbone)
        # self.ema_classifier_head.update_attr(self.classifier_head)
        # self.backbone = self.ema_backbone.ema
        # self.classifier_head = self.ema_classifier_head.ema
        if self.train_config.problem_type == "single_label_classification":
            total_correct_pred = sum(
                list(map(lambda x: x["correct_pred"], self.train_step_outputs))
            )
            total_pred_count = sum(
                list(map(lambda x: x["total_pred"], self.train_step_outputs))
            )
            train_acc = total_correct_pred / total_pred_count
            # For multi gpu training
            train_acc = self.all_gather(train_acc).mean().item()
            self.log("train/acc", train_acc, sync_dist=True)
            if self.global_rank == 0:
                print(f"\ntrain accuracy {train_acc} at epoch {self.current_epoch}")
            self.train_step_outputs.clear()

        # Multi-label classification
        else:
            TP = torch.sum(
                torch.stack(list(map(lambda x: x["TP"], self.train_step_outputs)))
            )
            FP = torch.sum(
                torch.stack(list(map(lambda x: x["FP"], self.train_step_outputs)))
            )
            FN = torch.sum(
                torch.stack(list(map(lambda x: x["FN"], self.train_step_outputs)))
            )
            TN = torch.sum(
                torch.stack(list(map(lambda x: x["TN"], self.train_step_outputs)))
            )

            train_precision = torch.nanmean(TP / (TP + FP))
            train_recall = torch.nanmean(TP / (TP + FN))
            train_f1 = (
                2 * train_precision * train_recall / (train_precision + train_recall)
            )
            # For multi gpu training
            train_precision = self.all_gather(train_precision).mean().item()
            train_recall = self.all_gather(train_recall).mean().item()
            train_f1 = self.all_gather(train_f1).mean().item()

            self.log("train/precision", train_precision, sync_dist=True)
            self.log("train/recall", train_recall, sync_dist=True)
            self.log("train/f1", train_f1, sync_dist=True)

            if self.global_rank == 0:
                print(
                    f"\ntrain precision {train_precision} at epoch {self.current_epoch}"
                )
                print(f"\ntrain recall {train_recall} at epoch {self.current_epoch}")
                print(f"\ntrain f1 {train_f1} at epoch {self.current_epoch}")

            self.train_step_outputs.clear()

        self.backbone = self.ema_backbone.ema[0]
        self.classifier_head = self.ema_classifier_head.ema[0]

        del self.ema_backbone
        del self.ema_classifier_head
        del self.backbone_list
        self.ema_backbone = None
        self.ema_classifier_head = None

    def on_validation_epoch_end(self):
        """
        Called at the end of validation epoch, here we log the validation accuracy or other metrics e.g. f1 score
        """
        if self.train_config.problem_type == "single_label_classification":
            # Print top k accuracy
            total_correct_pred_list = list(
                map(lambda x: x["correct_pred_list"], self.validation_step_outputs)
            )

            topk_correct_count_list = []
            for topk_idx in range(len(total_correct_pred_list[0])):
                topk_correct_count_list.append(
                    sum(list(map(lambda x: x[topk_idx], total_correct_pred_list)))
                )

            total_pred_count = sum(
                list(map(lambda x: x["total_pred"], self.validation_step_outputs))
            )

            topk_val_acc_list = []
            for topk_correct_count in topk_correct_count_list:
                topk_val_acc = topk_correct_count / total_pred_count
                # For multi gpu training
                topk_val_acc = self.all_gather(topk_val_acc).mean().item()
                topk_val_acc_list.append(topk_val_acc)

            for topk_val_acc, topk in zip(
                topk_val_acc_list, self.train_config.print_topk_list
            ):
                if self.global_rank == 0:
                    print(
                        f"\nTop {topk} Validation accuracy {topk_val_acc} at epoch {self.current_epoch}"
                    )
                self.log(f"val/acc{topk}", topk_val_acc, sync_dist=True)

            # Monitoring accuracy, used in naming checkpoint file.
            if self.train_config.monitor_topk != "total":
                monitor_topk_index = self.train_config.print_topk_list.index(
                    self.train_config.monitor_topk
                )
                self.log(
                    "val/monitor_acc",
                    topk_val_acc_list[monitor_topk_index],
                    sync_dist=True,
                )
            else:
                self.log("val/monitor_acc", sum(topk_val_acc_list), sync_dist=True)

            # Total sum accuracy
            self.log("val/tot_acc", sum(topk_val_acc_list), sync_dist=True)
            self.validation_step_outputs.clear()
        # Multi-label classification
        else:
            TP = torch.sum(
                torch.stack(list(map(lambda x: x["TP"], self.validation_step_outputs)))
            )
            FP = torch.sum(
                torch.stack(list(map(lambda x: x["FP"], self.validation_step_outputs)))
            )
            FN = torch.sum(
                torch.stack(list(map(lambda x: x["FN"], self.validation_step_outputs)))
            )
            TN = torch.sum(
                torch.stack(list(map(lambda x: x["TN"], self.validation_step_outputs)))
            )

            val_precision = torch.nanmean(TP / (TP + FP))
            val_recall = torch.nanmean(TP / (TP + FN))
            val_f1 = 2 * val_precision * val_recall / (val_precision + val_recall)
            # For multi gpu training
            val_precision = self.all_gather(val_precision).mean().item()
            val_recall = self.all_gather(val_recall).mean().item()
            val_f1 = self.all_gather(val_f1).mean().item()

            self.log("val/precision", val_precision, sync_dist=True)
            self.log("val/recall", val_recall, sync_dist=True)
            self.log("val/f1", val_f1, sync_dist=True)

            if self.global_rank == 0:
                print(f"\nval precision {val_precision} at epoch {self.current_epoch}")
                print(f"\nval recall {val_recall} at epoch {self.current_epoch}")
                print(f"\nval f1 {val_f1} at epoch {self.current_epoch}")

            self.validation_step_outputs.clear()

    def on_test_epoch_end(self):
        """
        Called at the end of test epoch, here we log the test accuracy or other metrics e.g. f1 score
        """
        if self.train_config.problem_type == "single_label_classification":
            # Print top k accuracy
            total_correct_pred_list = list(
                map(lambda x: x["correct_pred_list"], self.test_step_outputs)
            )

            topk_correct_count_list = []
            for topk_idx in range(len(total_correct_pred_list[0])):
                topk_correct_count_list.append(
                    sum(list(map(lambda x: x[topk_idx], total_correct_pred_list)))
                )

            total_pred_count = sum(
                list(map(lambda x: x["total_pred"], self.test_step_outputs))
            )

            topk_test_acc_list = []
            for topk_correct_count in topk_correct_count_list:
                topk_test_acc = topk_correct_count / total_pred_count
                # For multi gpu training
                topk_test_acc = self.all_gather(topk_test_acc).mean().item()
                topk_test_acc_list.append(topk_test_acc)

            for topk_test_acc, topk in zip(
                topk_test_acc_list, self.train_config.print_topk_list
            ):
                if self.global_rank == 0:
                    print(
                        f"\nTop {topk} Test accuracy {topk_test_acc} at epoch {self.current_epoch}"
                    )
                self.log(f"test/acc{topk}", topk_test_acc, sync_dist=True)

            # Total sum accuracy
            self.log("test/tot_acc", sum(topk_test_acc_list), sync_dist=True)
            self.test_step_outputs.clear()
        # Multi-label classification
        else:
            TP = torch.sum(
                torch.stack(list(map(lambda x: x["TP"], self.test_step_outputs)))
            )
            FP = torch.sum(
                torch.stack(list(map(lambda x: x["FP"], self.test_step_outputs)))
            )
            FN = torch.sum(
                torch.stack(list(map(lambda x: x["FN"], self.test_step_outputs)))
            )
            TN = torch.sum(
                torch.stack(list(map(lambda x: x["TN"], self.test_step_outputs)))
            )

            test_precision = torch.nanmean(TP / (TP + FP))
            test_recall = torch.nanmean(TP / (TP + FN))
            test_f1 = 2 * test_precision * test_recall / (test_precision + test_recall)
            # For multi gpu training
            test_precision = self.all_gather(test_precision).mean().item()
            test_recall = self.all_gather(test_recall).mean().item()
            test_f1 = self.all_gather(test_f1).mean().item()

            self.log("test/precision", test_precision, sync_dist=True)
            self.log("test/recall", test_recall, sync_dist=True)
            self.log("test/f1", test_f1, sync_dist=True)

            if self.global_rank == 0:
                print(
                    f"\ntest precision {test_precision} at epoch {self.current_epoch}"
                )
                print(f"\ntest recall {test_recall} at epoch {self.current_epoch}")
                print(f"\ntest f1 {test_f1} at epoch {self.current_epoch}")

            self.test_step_outputs.clear()

    def classifier_head_optimizers(self):
        """
        Returns the optimizer for the classifier head
        """
        return torch.optim.AdamW(
            self.classifier_head.parameters(),
            lr=self.train_config.classifier_learning_rate,
        )

    # Text Optimizer setup
    def add_optimizers(
        self,
        optimizer: torch.optim.Optimizer,
        backbone: nn.Module,
        freeze: bool,
        lora: bool,
        backbone_learning_rate: float,
    ):
        """
        Adds the optimizer for the backbone
        """
        # Classifier head fine-tuning
        if freeze:
            return optimizer
        # LORA fine-tuning
        elif lora:
            lora_pg = []
            for name, params in backbone.named_parameters():
                if "lora" in name:
                    lora_pg.append(params)
            optimizer.add_param_group({"params": lora_pg, "lr": backbone_learning_rate})
        # Full fine-tuning
        else:
            backbone_pg = []
            for name, params in backbone.named_parameters():
                backbone_pg.append(params)
            optimizer.add_param_group(
                {"params": backbone_pg, "lr": backbone_learning_rate}
            )
        return optimizer

    def get_lr_scheduler(self, optimizer, lr_scheduler_type: str):
        """
        Get the learning rate scheduler (Currently supporting only linear warmup, will be adding more in future)
        """
        if lr_scheduler_type == "linear_warmup":
            lr_scheduler = get_linear_schedule_with_warmup(
                optimizer=optimizer,
                num_warmup_steps=0.06 * self.trainer.estimated_stepping_batches,
                num_training_steps=self.trainer.estimated_stepping_batches,
            )
        elif lr_scheduler_type == "constant":
            lambda_constant = lambda epoch: 1.0
            lr_scheduler = torch.optim.lr_scheduler.LambdaLR(
                optimizer, lr_lambda=lambda_constant
            )
        else:
            raise ValueError(f"lr_scheduler_type {lr_scheduler_type} is not supported.")

        return lr_scheduler



import torch.nn as nn
import torch.nn.functional as F


class Base2LinearClassifier(nn.Module):
    """
    Head Classifier with 2 Linear Layers and relu, dropout 0.5 in between
    """

    def __init__(self, embedding_size, num_class):
        super().__init__()
        self.fc_layer1 = nn.Linear(embedding_size, embedding_size)
        self.fc_layer2 = nn.Linear(embedding_size, num_class)
        self.bn1 = nn.BatchNorm1d(embedding_size)
        self.dropout = nn.Dropout(0.5)

    def forward(self, x):
        x = self.fc_layer1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.fc_layer2(x)

        return x


class BaseLinearClassifier(nn.Module):
    """
    Head Classifier with 1 Linear Layer
    """

    def __init__(self, embedding_size, num_class):
        super().__init__()
        self.fc_layer1 = nn.Linear(embedding_size, num_class)

    def forward(self, x):
        x = self.fc_layer1(x)

        return x


class ElectraClassifier(nn.Module):
    """
    Classifier for Electra Model (following huggingface)
    """

    def __init__(self, embedding_size, num_class):
        super().__init__()
        self.fc_layer1 = nn.Linear(embedding_size, embedding_size)
        self.dropout = nn.Dropout(0.1)
        self.fc_layer2 = nn.Linear(embedding_size, num_class)

    def forward(self, x):
        x = self.dropout(x)
        x = self.fc_layer1(x)
        x = F.gelu(x)
        x = self.dropout(x)
        x = self.fc_layer2(x)

        return x



from .core.classifier.text_classifier import (
    TextClassifier,
    TextClassifierConfig,
    TextClassifierModule,
    FeedbackTextClassifierModule,
    FeedbackTextClassifier
)

from .core.classifier.image_classifier import (
    ImageClassifier,
    ImageClassifierConfig,
    ImageClassifierModule,
    FeedbackImageClassifier,
    FeedbackImageClassifierModule,
)

from .core.classifier.image_text_classifier import (
    ImageTextClassifier,
    ImageTextClassifierConfig,
    ImageTextClassifierModule,
    FeedbackImageTextClassifier,
    FeedbackImageTextClassifierModule
)

from .core.classifier.zeroshot_classifier import (
    ZeroshotImageClassifier,
    ZeroshotImageClassifierConfig,
)

from .core.pipeline import pipeline

from .core.classifier import (
    export_onnx,
    model_save,
    export_onnx_quantization,
    export_onnx_optimization,
)



import torch
from ..head_classifier import *

from .image_text_config import ImageTextClassifierConfig
from ..base_trainer import BaseClassifierModule
from ..base_feedback_trainer import FeedbackBaseClassifierModule


class ImageTextClassifierModule(BaseClassifierModule):
    """
    Image Text Classifier Module.

    Args:
        train_config (ImageTextClassifierConfig): ImageTextClassifierConfig.
    """

    def __init__(self, train_config: ImageTextClassifierConfig):
        super().__init__(train_config)

        (
            self.image_backbone,
            self.train_transforms,
            self.val_transforms,
            self.image_data_config,
            self.train_config.image_timm,
            self.image_initial_config,
            self.text_backbone,
            self.tokenizer,
            self.text_initial_config,
            self.delete_input_key_list,
        ) = self.get_model()
        # Concat embedding
        self.image_embedding_size = self.image_backbone.forward(
            torch.randn(self.image_data_config["input_size"]).unsqueeze(0)
        ).shape[-1]
        self.text_embedding_size = self.text_initial_config.hidden_size
        self.embedding_size = self.image_embedding_size + self.text_embedding_size

        # Define classifier
        if type(self.train_config.fc_classifier) == str:
            classifier_class = globals()[self.train_config.fc_classifier]
            self.classifier_head = classifier_class(
                self.embedding_size, self.train_config.num_class
            )
        else:
            self.classifier_head = self.train_config.fc_classifier(
                self.embedding_size, self.train_config.num_class
            )

        self.save_hyperparameters()

    def get_model(self):
        """
        Get model from config.
        Return image_backbone, train_transforms, val_transforms, image_initial_config, image_timm, text_backbone, tokenizer, text_initial_config, delete_input_key_list
        """
        # Get Text Models
        get_text_model_result = self.get_text_model(
            self.train_config.text_backbone_name,
            self.train_config.text_freeze,
            self.train_config.text_lora,
            self.train_config.text_lora_r,
            self.train_config.text_lora_alpha,
            self.train_config.text_lora_target_modules,
            self.train_config.text_lora_dropout,
        )
        get_image_model_result = self.get_image_model(
            self.train_config.image_backbone_name,
            self.train_config.image_freeze,
            self.train_config.image_lora,
            self.train_config.image_lora_r,
            self.train_config.image_lora_alpha,
            self.train_config.image_lora_target_modules,
            self.train_config.image_lora_dropout,
        )

        return get_image_model_result + get_text_model_result

    # model forward
    def forward(
        self,
        input_img_paths,
        input_texts,
        image_backbone=None,
        text_backbone=None,
        classifier_head=None,
        eval=False,
        classifier=True,
    ):
        if image_backbone is None:
            image_backbone = self.image_backbone
        if text_backbone is None:
            text_backbone = self.text_backbone
        if classifier_head is None:
            classifier_head = self.classifier_head
        image_backbone_outputs = self.image_forward(
            input_img_paths,
            image_backbone,
            classifier_head,
            self.train_config.image_freeze,
            eval=eval,
            classifier=False,
        )
        text_backbone_outputs = self.text_forward(
            input_texts,
            text_backbone,
            classifier_head,
            self.text_initial_config,
            self.train_config.text_freeze,
            eval=eval,
            classifier=False,
        )

        backbone_outputs = torch.cat(
            [image_backbone_outputs, text_backbone_outputs], dim=1
        )
        if classifier:
            logits = classifier_head(backbone_outputs)

            return logits
        else:
            return backbone_outputs

    def configure_optimizers(self):
        optimizer = self.classifier_head_optimizers()
        optimizer = self.add_optimizers(
            optimizer,
            self.text_backbone,
            self.train_config.text_freeze,
            self.train_config.text_lora,
            self.train_config.text_backbone_learning_rate,
        )
        optimizer = self.add_optimizers(
            optimizer,
            self.image_backbone,
            self.train_config.image_freeze,
            self.train_config.image_lora,
            self.train_config.image_backbone_learning_rate,
        )

        lr_scheduler = self.get_lr_scheduler(optimizer, self.train_config.lr_scheduler)

        return [optimizer], [{"scheduler": lr_scheduler}]


class FeedbackImageTextClassifierModule(FeedbackBaseClassifierModule):
    """
    Feedback Image Text Classifier Module.

    Args:
        train_config (ImageTextClassifierConfig): ImageTextClassifierConfig.
    """

    def __init__(self, train_config: ImageTextClassifierConfig):
        super().__init__(train_config)

        (
            self.image_backbone,
            self.train_transforms,
            self.val_transforms,
            self.image_data_config,
            self.train_config.image_timm,
            self.image_initial_config,
            self.text_backbone,
            self.tokenizer,
            self.text_initial_config,
            self.delete_input_key_list,
        ) = self.get_model()
        # Concat embedding
        self.image_embedding_size = self.image_backbone.forward(
            torch.randn(self.image_data_config["input_size"]).unsqueeze(0)
        ).shape[-1]
        self.text_embedding_size = self.text_initial_config.hidden_size
        self.embedding_size = self.image_embedding_size + self.text_embedding_size

        # Define classifier
        if type(self.train_config.fc_classifier) == str:
            classifier_class = globals()[self.train_config.fc_classifier]
            self.classifier_head = classifier_class(
                self.embedding_size, self.train_config.num_class
            )
        else:
            self.classifier_head = self.train_config.fc_classifier(
                self.embedding_size, self.train_config.num_class
            )

        self.save_hyperparameters()

    def get_model(self):
        """
        Get model from config.
        Return image_backbone, train_transforms, val_transforms, image_initial_config, image_timm, text_backbone, tokenizer, text_initial_config, delete_input_key_list
        """
        # Get Text Models
        get_text_model_result = self.get_text_model(
            self.train_config.text_backbone_name,
            self.train_config.text_freeze,
            self.train_config.text_lora,
            self.train_config.text_lora_r,
            self.train_config.text_lora_alpha,
            self.train_config.text_lora_target_modules,
            self.train_config.text_lora_dropout,
        )
        get_image_model_result = self.get_image_model(
            self.train_config.image_backbone_name,
            self.train_config.image_freeze,
            self.train_config.image_lora,
            self.train_config.image_lora_r,
            self.train_config.image_lora_alpha,
            self.train_config.image_lora_target_modules,
            self.train_config.image_lora_dropout,
        )

        return get_image_model_result + get_text_model_result

    # model forward
    def forward(
        self,
        input_img_paths,
        input_texts,
        image_backbone=None,
        text_backbone=None,
        classifier_head=None,
        eval=False,
        classifier=True,
    ):
        if image_backbone is None:
            image_backbone = self.image_backbone
        if text_backbone is None:
            text_backbone = self.text_backbone
        if classifier_head is None:
            classifier_head = self.classifier_head
        image_backbone_outputs = self.image_forward(
            input_img_paths,
            image_backbone,
            classifier_head,
            self.train_config.image_freeze,
            eval=eval,
            classifier=False,
        )
        text_backbone_outputs = self.text_forward(
            input_texts,
            text_backbone,
            classifier_head,
            self.text_initial_config,
            self.train_config.text_freeze,
            eval=eval,
            classifier=False,
        )

        backbone_outputs = torch.cat(
            [image_backbone_outputs, text_backbone_outputs], dim=1
        )
        if classifier:
            logits = classifier_head(backbone_outputs)

            return logits
        else:
            return backbone_outputs

    def configure_optimizers(self):
        optimizer = self.classifier_head_optimizers()
        optimizer = self.add_optimizers(
            optimizer,
            self.text_backbone,
            self.train_config.text_freeze,
            self.train_config.text_lora,
            self.train_config.text_backbone_learning_rate,
        )
        optimizer = self.add_optimizers(
            optimizer,
            self.image_backbone,
            self.train_config.image_freeze,
            self.train_config.image_lora,
            self.train_config.image_backbone_learning_rate,
        )

        lr_scheduler = self.get_lr_scheduler(optimizer, self.train_config.lr_scheduler)

        return [optimizer], [{"scheduler": lr_scheduler}]



import torch
import logging

logger = logging.getLogger(__name__)


class ZeroshotImageClassifierConfig:
    """
    Zeroshot Image Classifier Config

    Args:
        backbone_name (str): backbone name
        model_name (str): model name
    """

    def __init__(
        self,
        backbone_name: str = "clip",
        model_name: str = "patrickjohncyh/fashion-clip",
    ):
        self.device = "cuda"

        self.num_class = 0
        self.output_dir = "tmp_classifier"
        self.num_workers = 4
        self.text_prompt = "This is a photo of {}"

        self.backbone_name = backbone_name  # "clip"
        self.model_name = model_name  # "patrickjohncyh/fashion-clip"
        self.embedding_size = 512

        self.print_topk_list = [1, 3, 5]
        self.label_name_list = None



from setuptools import setup, find_packages


def read_requirements():
    with open("requirements.txt", "r") as req:
        content = req.read()
        requirements = content.split("\n")
    return requirements


setup(
    name="dalpha_ai",
    version="0.3.4",
    packages=find_packages(),
    install_requires=read_requirements(),
)



import pandas as pd
from torch.utils.data import Dataset
from typing import List
import re


class TextDataset(Dataset):
    def __init__(self, text_list: List[str], text_prompt: str):
        """
        Text Dataset

        Args:
            text_list (List[str]): list of text input
            text_prompt (str): text prompt
        """
        super().__init__()
        self.text_list = text_list
        self.text_prompt = text_prompt

    def __getitem__(self, idx):
        text = self.text_list[idx]
        if bool(re.match("^[a-zA-Z\s><.,!?|]*$", text)):
            text = self.text_prompt.format(text)
        else:
            raise ValueError(
                f"Failed to create embedding vectors from text '{text}', The text must be composed only of English letters and ' ' ',' '.' '|' '>' '<' '!' '?'."
            )

        return text

    def __len__(self):
        return len(self.text_list)


class ImageDataset(Dataset):
    """
    Image Dataset

    Args:
        img_path_list (List[str]): list of image path
    """

    def __init__(self, img_path_list):
        super().__init__()
        self.img_path_list = img_path_list

    def __getitem__(self, idx):
        input_img_path = self.img_path_list[idx]

        return input_img_path

    def __len__(self):
        return len(self.img_path_list)


class ImageClassifierDataset(Dataset):
    """
    Image Classifier Dataset

    Args:
        data (pd.DataFrame): dataframe of image path and label
    """

    def __init__(self, data: pd.DataFrame):
        super().__init__()
        self.img_path_list = data["img_path"].tolist()
        self.label_list = data["label"].tolist()

    def __getitem__(self, idx):
        input_img_path = self.img_path_list[idx]
        label = self.label_list[idx]

        return input_img_path, label

    def __len__(self):
        return len(self.label_list)
